////////////////////////////////////////////////////////////////////////////////
//! \file
//!
//! \author	Mauro Fiorentini
//!
//! \date	12/09/2021
//!
//! \par	Function:
//!	Contains the declaration of the GraphicSpace::Bitmap class.
//!
////////////////////////////////////////////////////////////////////////////////

# if	!defined BITMAP_H

# define	BITMAP_H

# include	"RandomGenerator.h"
# include	"Mutex.h"

# include	<stddef.h>
# include	<stdlib.h>

namespace	GraphicSpace
{

////////////////////////////////////////////////////////////////////////////////
//! \class	Bitmap
//!	The Bitmap class represents an image in bitmap format.
//!
////////////////////////////////////////////////////////////////////////////////

class	Bitmap
	{
	public:

//! \property	LONG_DATA_SIZE
//!	Size of a 16 bit data field, in bytes.
		static	constexpr	unsigned int	SHORT_DATA_SIZE = 2;

//! \property	LONG_DATA_SIZE
//!	Size of a 32 bit data field, in bytes.
		static	constexpr	unsigned int	LONG_DATA_SIZE = 4;

//! \property	HEADER_SIZE
//!	Size of the bitmap file header in bytes.
		static	constexpr	unsigned int	HEADER_SIZE = 14;

//! \property	HEADER_NAME_LENGTH	
//!	Length of the name of the file type, in bytes
		static	constexpr	unsigned int	HEADER_NAME_LENGTH = 2;

//! \property	OFFSET_OF_BITMAP_OFFSET
//!	Offset of the bitmap offset within the header.
		static	constexpr	unsigned int
						OFFSET_OF_BITMAP_OFFSET = 10;

//! \property	OFFSET_OF_FILE_SIZE
//!	Offset of the file size within the header.
		static	constexpr	unsigned int	OFFSET_OF_FILE_SIZE = 2;

//! \property	DIB_HEADER_SIZE
//!	Size of the bitmap file DIB header in bytes.
		static	constexpr	unsigned int	DIB_HEADER_SIZE = 40;

//! \property	OFFSET_OF_WIDTH
//!	Offset of the image width within the DIB header.
		static	constexpr	unsigned int	OFFSET_OF_WIDTH = 4;

//! \property	OFFSET_OF_HEIGHT
//!	Offset of the image height within the DIB header.
		static	constexpr	unsigned int	OFFSET_OF_HEIGHT = 8;

//! \property	OFFSET_OF_COLOR_PLANES
//!	Offset of the number of color planes within the DIB header.
		static	constexpr	unsigned int	OFFSET_OF_COLOR_PLANES =
								12;

//! \property	OFFSET_OF_BITS_PER_PIXEL
//!	Offset of the number of bits per pixel within the DIB header.
		static	constexpr	unsigned int
						OFFSET_OF_BITS_PER_PIXEL = 14;

//! \property	OFFSET_OF_COMPRESSION
//!	Offset of the kind of compression within the DIB header.
		static	constexpr	unsigned int	OFFSET_OF_COMPRESSION =
								16;
//! \property	OFFSET_OF_IMAGE_SIZE
//!	Offset of the image size within the DIB header.
		static	constexpr	unsigned int	OFFSET_OF_IMAGE_SIZE =
								20;

//! \property	OFFSET_OF_HORIZONTAL_RESOLUTION
//!	Offset of the horizontal resolution within the DIB header.
		static	constexpr	unsigned int
					OFFSET_OF_HORIZONTAL_RESOLUTION = 24;

//! \property	OFFSET_OF_VERTICAL_RESOLUTION
//!	Offset of the vertical resolution within the DIB header.
		static	constexpr	unsigned int
					OFFSET_OF_VERTICAL_RESOLUTION = 28;

//! \property	OFFSET_OF_NUMBER_OF_COLORS
//!	Offset of the number of colors within the DIB header.
		static	constexpr	unsigned int
					OFFSET_OF_NUMBER_OF_COLORS = 32;

//! \property	OFFSET_OF_NUMBER_OF_IMPORTANT_COLORS
//!	Offset of the number of important colors within the DIB header.
		static	constexpr	unsigned int
				OFFSET_OF_NUMBER_OF_IMPORTANT_COLORS = 36;

//! \property	NUMBER_OF_COLORS
//!	Number of color planes.
		static	constexpr	unsigned int	NUMBER_OF_COLORS = 3;

//! \property	MAXIMUM_WIDTH
//!	Maximum width of an image.
		static	constexpr	unsigned int	MAXIMUM_WIDTH = 10000;

//! \property	MAXIMUM_HEIGHT
//!	Maximum Height of an image.
		static	constexpr	unsigned int	MAXIMUM_HEIGHT = 10000;

//! \property	MAXIMUM_BITMAP_SIZE
//!	Maximum number of pixel in an image.
		static	constexpr	unsigned int	MAXIMUM_BITMAP_SIZE =
								20000000;

//! \property	NUMBER_OF_COLOR_PLANES
//!	Valid number of color planes.
		static	constexpr	unsigned int	NUMBER_OF_COLOR_PLANES =
								1;
//! \property	VALID_COMPRESSION
//!	Valid value for compression.
		static	constexpr	unsigned int	VALID_COMPRESSION = 0;

//! \property	MAXIMUM_NUMBER_OF_THREADS
//!	Maximum number of threads.
		static	constexpr	unsigned int
						MAXIMUM_NUMBER_OF_THREADS = 100;

//! \property	MAXIMUM_NUMBER_OF_TERMITES
//!	Maximum number of termites.
		static	constexpr	unsigned int
						MAXIMUM_NUMBER_OF_TERMITES =
								1000;

//! \property	MAXIMUM_NUMBER_OF_STEPS
//!	Maximum number of steps.
		static	constexpr	unsigned int
						MAXIMUM_NUMBER_OF_STEPS = 1000;

//! \typedef	IndexType
//!	The type used as pixel index.
		typedef	uint_fast32_t	IndexType;

////////////////////////////////////////////////////////////////////////////////
//! \enum	ChoiceType
//!	The ChoiceType enumerated type specify how the choiche of the step is
//!	done.
//!
////////////////////////////////////////////////////////////////////////////////

		enum	class	ChoiceType
			{
				//! The step is randomly chosen with a
				//! probability distribution
			RANDOM,

				//! The step with maximum probability is chosen.
			MAXIMUM

			};

////////////////////////////////////////////////////////////////////////////////
//! \struct	CorrectionParameters
//!	The CorrectionParameters struct contains the parameters for the
//!	correction of an image.
//!
////////////////////////////////////////////////////////////////////////////////

		struct	CorrectionParameters
			{

//! \property	Threads
//!	Number of threads.
			unsigned int	Threads;

//! \property	Termites
//!	Number of termites.
			unsigned int	Termites;

//! \property	Steps
//!	Number of steps.
			unsigned int	Steps;

//! \property	RepulsionStep
//!	Value of the repulsion step parameter.
			float		RepulsionStep;

//! \property	Alpha
//!	Exponent of the repulsion parameter.
			float		Alpha;

//! \property	Beta
//!	Exponent of the attraction parameter.
			float		Beta;

//! \property	RandomizeFlag
//!	Flag specifying whether the random number genarator shall be initialized
//!	with a not reproducible seed.
			bool		RandomizeFlag;

//! \property	ChoiceKind
//!	Specification of how the choice of the step is done.
			ChoiceType	ChoiceKind;

//! \property	TruncatePath
//!	Flag indicating whether a path shall be truncated when the a pixel with
//!	the maximum value of the color is reached.
			bool		TruncatePath;

//! \property	Quiet
//!	Flag indicating whether the program shall not print information about
//!	the image.
			bool		Quiet;

			};

	private:

//! \property	STRIP_INDEX_FORMAT 
//!	Format used to print the index of the strip being processed.
		static	constexpr	const	char*	STRIP_INDEX_FORMAT =
					"Processing strip %" PRIuFAST32 "\n";

//! \property	INITIAL_GENERATION
//!	Number of random numbers generated during the initialization of the
//!	generator.
		static	constexpr	unsigned int	INITIAL_GENERATION =
								1000;

//! \property	NUMBER_OF_NEIGHBOURS
//!	Number of neigbours of every pixel.
		static	constexpr	unsigned int	NUMBER_OF_NEIGHBOURS =
								8;
//! \property	ATTRACTION_SCALE
//!	Scale factor for attraction.
//!
//! \note
//!	The product ATTRACTION_SCALE * REPULSION_SCALE shall not exceed
//!	0x200000.
//!
//! \see	REPULSION_SCALE
		static	constexpr	unsigned int	ATTRACTION_SCALE =
								0x800;

//! \property	REPULSION_SCALE
//!	Scale factor for repulsion.
//!
//! \note
//!	REPULSION_SCALE shall not exceed USHRT_MAX and the product
//!	ATTRACTION_SCALE * REPULSION_SCALE shall not exceed 0x200000.
//!
//! \see	ATTRACTION_SCALE
		static	constexpr	unsigned int	REPULSION_SCALE = 0x400;

//! \property	TERMITES_SUM_SCALE
//!	Scale factor for the sum of the contribution of the termites.
//!
//! \note
//!	The product UINT8_MAX * MAXIMUM_NUMBER_OF_TERMITES * TERMITES_SUM_SCALE
//!	shall not exceed UINT32_MAX.
//!
//! \see	MAXIMUM_NUMBER_OF_TERMITES
		static	constexpr	unsigned int	TERMITES_SUM_SCALE =
								0x4000;

//! \property	REPULSION_TABLE_SIZE
//!	Number of elements in the repulsion table.
//!
//! \see	MAXIMUM_NUMBER_OF_TERMITES
//! \see	MAXIMUM_NUMBER_OF_STEPS
		static	constexpr	size_t	REPULSION_TABLE_SIZE =
						static_cast<size_t>
						(MAXIMUM_NUMBER_OF_TERMITES) *
						static_cast<size_t>
						(MAXIMUM_NUMBER_OF_STEPS) / 2;

//! \property	STRIP_HEIGHT
//!	Height of the strips given to each thread.
		static	constexpr	unsigned int	STRIP_HEIGHT = 5;

////////////////////////////////////////////////////////////////////////////////
//! \struct	RepulsionType
//!	The RepulsionType struct contains the information about the repulsion of
//!	a pixel.
//!
////////////////////////////////////////////////////////////////////////////////

		struct	RepulsionType
			{

//! property	Repulsion
//!	Repulsion value.
//!
//! \note
//!	The value is the floating point value, multiplied for REPULSION_SCALE
//!	and rounded to the nearest integer.
			unsigned short	Repulsion;

//! property	Visitors
//!	Number of termites which already visited the pixel.
			unsigned short	Visitors;

			};

////////////////////////////////////////////////////////////////////////////////
//! \struct	PixelInformationType
//!	The PixelInformationType struct contains the information related to a
//!	pixel.
//!
////////////////////////////////////////////////////////////////////////////////

		struct	PixelInformationType
			{

////////////////////////////////////////////////////////////////////////////////
//! \struct	NeighbourType
//!	The NeighbourType struct contains the information about a neighbour
//!	pixel.
//!
////////////////////////////////////////////////////////////////////////////////

			struct	NeighbourType
				{

//! property	Index
//!	Index of the neighbour pixel.
				uint_least32_t	Index;

//! property	Attraction
//!	Value of the attraction between the pixel and its neighbour.
//!
//! \note
//!	The value is the floating point value, multiplied for ATTRACTION_SCALE
//!	and rounded to the nearest integer.
				uint_least32_t	Attraction;

				bool		SetLink(const Bitmap& Image,
							unsigned int Row,
							unsigned int Colum)
							noexcept;

				};
		
//! property	NumberOfNeighbours
//!	Number of neigbour pixels.
			unsigned int	NumberOfNeighbours;

//! property	Neigbour
//!	Array of information about neighbour pixels.
//!
//! \see	NUMBER_OF_NEIGHBOURS
			NeighbourType	Neighbour [NUMBER_OF_NEIGHBOURS];

			void		SetLinks(const Bitmap& Image,
						unsigned int Row,
						unsigned int Colum) noexcept;

			void		SetAttraction(register const Bitmap&
						Image, register const uint8_t*
						ColorPointer, register
						IndexType PixelIndex,
						register const uint_least32_t
						OrthogonalAttraction
						[UINT8_MAX + 1],
						register const uint_least32_t
						DiagonalAttraction
						[UINT8_MAX + 1]) noexcept;

			IndexType	NextPixelMaximum(RandomGenerator&
						CurrentGenerator, register const
						RepulsionType* RepulsionPointer)
						const noexcept;

			IndexType	NextPixelRandom(RandomGenerator&
						CurrentGenerator, register const
						RepulsionType* RepulsionPointer)
						const noexcept;

			};

//! \typedef	PathFunction
//!	Pointer to the function which chooses the next pixel.
//!
//! \note
//!	The function captures the random number generator and the pointer to the
//!	array of RepulsionType.
		typedef	IndexType	(PixelInformationType::*PathFunction)
						(RandomGenerator&
						CurrentGenerator, const
						RepulsionType* RepulsionPointer)
						const;

////////////////////////////////////////////////////////////////////////////////
//! \struct	CorrectionDescriptorType
//!	The CorrectionDescriptor struct contains the parameters for the
//!	correction of a strip of an image.
//!
////////////////////////////////////////////////////////////////////////////////

		struct	CorrectionDescriptorType
			{

//! \param	StartRow
//!	Index of the first row of the strip.
				unsigned int	StartRow;

//! \param	EndRow
//!	Index of the first row after the strip.
				unsigned int	EndRow;

//! \param	CorrectedColorPointer
//!	Pointer to the array that will contain the corrected bitmap of the
//!	color.
				uint8_t*	CorrectedColorPointer;

//! \param	OriginalColorPointer
//!	Pointer to the array that contains the the original bitmap of the color.
			const	uint8_t*	OriginalColorPointer;

//! \param	MaximumValue
//!	Maximum value of the color in the whole image.
				unsigned int	MaximumValue;

//! \param	Parameters
//!	Struct containing the correction parameters.
			const 	CorrectionParameters*
						ParametersPointer;

//! \param	PixelInformationPointer
//!	Pointer to the array of information about the pixel.
			const	PixelInformationType*
						PixelInformationPointer;

//! \param	PathFinder
//!	Member pointer to the function used to select the pixels to visit.
				PathFunction	PathFinder;

//! \param	DivisionTablePointer
//!	Pointer to the array of quotient: for n > 0 the entry with index n
//!	contains UINT8_MAX * TERMITES_SUM_SCALE / n, rounded to the closes
//!	integer; the entry with index 0 contains the same value of the entry
//!	with index 1.
			const	uint_least32_t*	DivisionTablePointer;

			};

//! \property	Width
//!	Width of the image in pixels.
		uint_fast32_t	Width;

//! \property	Height
//!	Height of the image in pixels.
		uint_fast32_t	Height;

//! \property	BitmapSize
//!	Size of the image in pixels.
		uint_fast32_t	BitmapSize;

//! \property	BitsPerPixel
//!	Bits per pixels.
		uint16_t	BitsPerPixel;

//! \property	HorizontalResolution
//!	Horizontal resolution in pixel per meter.
		uint_fast32_t	HorizontalResolution;

//! \property	VerticalResolution
//!	Vertical resolution in pixel per meter.
		uint_fast32_t	VerticalResolution;

//! \property	PixelPointer
//!	Array of pointers to bitmaps, one for each color.
		uint8_t*	PixelPointer [NUMBER_OF_COLORS];

//! property	RepulsionTable
//!	Array of repulsion values; the index is the number of termites which
//!	already visited a pixel.
//!
//! \note
//!	It is declared mutable, as it is necessary to change its content to
//!	correct the image, even if the Bitmap is not changed.
//!
//! \see	REPULSION_TABLE_SIZE
		mutable	unsigned short	RepulsionTable [REPULSION_TABLE_SIZE];

//! property	InitialRepulsionPointer
//!	Pointer to an array of initial repulsion values, corresponding to zero
//!	visitors.
//!
//! \note
//!	It is declared mutable, as it is necessary to change its content to
//!	correct the image, even if the Bitmap is not changed.
		mutable	RepulsionType*	InitialRepulsionPointer;

//! \property	Generator
//!	Random numbers generator.
//!
//! \note
//!	It is declared mutable, as it is necessary to change its content to
//!	generate new numbers, even if the Bitmap is not changed.
		mutable	RandomGenerator	Generator;

//! \property	Strips
//!	Number of strips.
		uint_fast32_t		Strips;

//! \property	StripMutex
//!	Mutex used by the thread to get new strips to process.
		Mutex			StripMutex;

//! \property	DescriptorPointer
//!	Array of descriptors used to split the work among threads.
		CorrectionDescriptorType*	DescriptorPointer;

//! \property	HeaderName
//!	Name of the file type.
		static	const	char	HeaderName [HEADER_NAME_LENGTH];

//! \property	OpenReadBinary
//!	String used to specify binary read only mode when opening a file.
		static	const	char*	const	OpenReadBinary;

//! \property	OpenWriteBinary
//!	String used to specify binary write only mode when opening a file.
		static	const	char*	const	OpenWriteBinary;

			void		AllocateImage(void) noexcept;

			void		DestroyImage(void) noexcept;

			bool		ValidIndex(unsigned int Row,
						unsigned int Column) const
						noexcept;

			IndexType	MakeIndex(unsigned int Row,
						unsigned int Column) const
						noexcept;

			bool		IsOrthogonal(IndexType FirstIndex,
						IndexType SecondIndex) const
						noexcept;

			unsigned int	ComputeTermite(register unsigned int
						MaximumValue,
						register IndexType PixelIndex,
						register const uint8_t*
						ColorPointer,
						register unsigned int Steps,
						register const
						PixelInformationType*
						PixelInformationPointer,
						register RepulsionType*
						RepulsionPointer,
						register RandomGenerator&
						CurrentGenerator, register
						PathFunction PathFinder) const
						noexcept;

			void		CorrectStrip(register uint8_t*
						CorrectedColorPointer,
						register const uint8_t*
						OriginalColorPointer,
						register unsigned int StartRow,
						register unsigned int EndRow,
						register unsigned int
						MaximumValue,
						register const
						CorrectionParameters&
						Parameters,
						register const
						PixelInformationType*
						PixelInformationPointer,
						register PathFunction
						PathFinder,
						register const uint_least32_t
						DivisionTable [UINT8_MAX + 1])
						const noexcept;

			void		CorrectColor(register uint8_t*
						CorrectedColorPointer,
						register const uint8_t*
						OriginalColorPointer,
						register unsigned int
						MaximumValue,
						register const
						CorrectionParameters&
						Parameters,
						register const
						PixelInformationType*
						PixelInformationPointer,
						register PathFunction
						PathFinder,
						const uint_least32_t
						DivisionTable [UINT8_MAX + 1])
						noexcept;

		const	CorrectionDescriptorType*
					GetStripDescriptor(void) noexcept;

		static	uint_fast32_t	RoundedDivide(uint_fast32_t Dividend,
						uint_fast32_t Divisor) noexcept;

		static	uint_fast32_t	LoadInteger(const uint8_t Buffer [],
						unsigned int Size) noexcept;

		static	void		StoreInteger(uint8_t Buffer [],
						uint_fast32_t Value,
						unsigned int Size) noexcept;

	public:

				Bitmap(void) noexcept;

				Bitmap(const Bitmap& thisBitmap) noexcept;

				Bitmap(Bitmap&& thisBitmap) noexcept;

				~Bitmap(void) noexcept;

		Bitmap&		operator = (const Bitmap& thisBitmap) noexcept;

		Bitmap&		operator = (Bitmap&& thisBitmap) noexcept;

		uint_fast32_t	GetWidth(void) const noexcept;

		uint_fast32_t	GetHeight(void) const noexcept;

		uint_fast32_t	GetSize(void) const noexcept;

		unsigned int	GetBitsPerPixel(void) const noexcept;

		uint_fast32_t	GetHorizontalResolution(void) const noexcept;

		uint_fast32_t	GetVerticalResolution(void) const noexcept;

		uint_fast32_t	ComputeLineSize(void) const noexcept;

		void		ReadFile(const char* Path) noexcept(false);

		void		WriteFile(const char* Path) noexcept(false);

		void		Randomize(uint_fast32_t Seed [RandomGenerator::
					NUMBER_OF_GENERATORS + 1]) noexcept;

		Bitmap		Correct(register const CorrectionParameters&
					Parameters) noexcept(false);

		static	void*	CorrectFunction(void* Argument) noexcept(false);

	};

}

# include	"Bitmap.inl"

# endif
