////////////////////////////////////////////////////////////////////////////////
//! \file
//!
//! \author	Mauro Fiorentini
//!
//! \date	20/10/2021
//!
//! \par	Function:
//!	Contains the declaration of the
//!	GraphicSpace::EnumeratedParameterDescriptor class template.
//!
////////////////////////////////////////////////////////////////////////////////

# if	!defined ENUMERATED_PARAMETER_DESCRIPTOR_H

# define	ENUMERATED_PARAMETER_DESCRIPTOR_H

# include	"Bitmap.h"
# include	"ParameterDescriptor.h"
# include	"EnumeratedException.h"
# include	"FileException.h"

namespace	GraphicSpace
{

////////////////////////////////////////////////////////////////////////////////
//! \class	EnumeratedParameterDescriptor
//!	The EnumeratedParameterDescriptor class template represents the
//!	descriptor of a command line parameter with enumerated value.
//!
//! \tparam	ENUMERATED_TYPE	The type of the parameter; it must be an
//!				enumerated type.
//!
////////////////////////////////////////////////////////////////////////////////

template<typename ENUMERATED_TYPE>
class	EnumeratedParameterDescriptor: public ParameterDescriptor
	{

	private:

//! \property	ParameterPointer
//!	Member pointer to the parameter within the CorrectionParameter struct.
		ENUMERATED_TYPE	Bitmap::CorrectionParameters::*
					ParameterPointer;

//! \property	ExceptionKind
//!	The kind of exception that is generated if an enumerated command line
//!	parameter is not correct.
		EnumeratedExceptionType	ExceptionKind;

//! \property	ValueString
//!	Array of strings used to specify the possible values of an enumerated
//!	parameter.
		static	const	char*	const	ValueString [];

//! \property	UsageFormatHead
//!	Format used to write the first part of the usage message related to an
//!	enumerated parameter.
		static	constexpr	const	char*	UsageFormatHead =
								"\t%s%c<";
//! \property	UsageFormatString
//!	Format used to write a value of an enumerated parameter.
		static	constexpr	const	char*	UsageFormatString =
								"%s";

//! \property	UsageFormatTail
//!	Format used to write the first part of the usage message related to an
//!	enumerated parameter.
		static	constexpr	const	char*	UsageFormatTail =
							"> (default %s)\n";

//! \property	SEPARATOR_CHARACTER
//!	Character used to separate the possible values of an enumerated
//!	parameter.
		static	constexpr	char		SEPARATOR_CHARACTER =
								'|';

				EnumeratedParameterDescriptor(const
					EnumeratedParameterDescriptor&
					Descriptor) noexcept = delete;

				EnumeratedParameterDescriptor
					(EnumeratedParameterDescriptor&&
					Descriptor) noexcept = delete;

		void		operator = (const EnumeratedParameterDescriptor&
					Descriptor) noexcept = delete;

		void		operator = (EnumeratedParameterDescriptor&&
					Descriptor) noexcept = delete;

		unsigned int	GetNumberOfValues(void) const noexcept;

	public:

				EnumeratedParameterDescriptor(const char*
					thisName, ENUMERATED_TYPE
					Bitmap::CorrectionParameters::*
					thisParameterPointer,
					EnumeratedExceptionType thisException)
					noexcept;

		virtual		~EnumeratedParameterDescriptor(void) noexcept =
					default;

			void	Assign(const char* String,
					Bitmap::CorrectionParameters&
					Parameters) const noexcept(false)
					override;

			void	PrintUsage(const Bitmap::CorrectionParameters&
					DefaultParameters) const noexcept(false)
					override;

	};

}

# include	"EnumeratedParameterDescriptor.tpl"

# endif
