////////////////////////////////////////////////////////////////////////////////
//! \file
//!
//! \author	Mauro Fiorentini
//!
//! \date	20/10/2021
//!
//! \par	Function:
//!	It contains the definitions of the functions of the inline functions of
//!	the GraphicSpace::EnumeratedParameterDescriptor class template.
//!
////////////////////////////////////////////////////////////////////////////////

namespace	GraphicSpace
{

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Constructor.
//!
//! \param	thisName		Pointer to the string containing the
//!					name of the parameter.
//! \param	thisParameterPointer	Member pointer to the parameter within
//!					the CorrectionParameter struct.
//! \param	thisException		The kind of exception that is generated
//!					if an enumerated command line parameter
//!					is not correct.
//!
//! \throw	None.
//!
//! \returns	Not applicable.
//!
//! \note
//!	if thisName is not a valid pointer to a properly terminated string, the
//!	behavior is undefined.
//!
////////////////////////////////////////////////////////////////////////////////

template<typename ENUMERATED_TYPE>
inline	EnumeratedParameterDescriptor<ENUMERATED_TYPE>::
		EnumeratedParameterDescriptor(const char* thisName,
		ENUMERATED_TYPE Bitmap::CorrectionParameters::*
		thisParameterPointer, EnumeratedExceptionType thisException)
		noexcept: ParameterDescriptor(thisName),
		ParameterPointer(thisParameterPointer),
		ExceptionKind(thisException)
	{
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Assign an enumerated value to a parameter.
//!
//! \param	String		Pointer to the string containing the value of
//!				the parameter.
//! \param	Parameters	Reference to the CorrectionParameter struct to
//!				modify with the new value.
//!
//! \throw	ExceptionKind	If String does not begin with an equal sign or
//!				String does not correspond to a enumerated
//!				value.
//!
//! \returns	None.
//!
//! \note
//!	if String is not a valid pointer to a properly terminated string, the
//!	behavior is undefined.
//!
//! \see	EQUAL_CHARACTER
//! \see	END_STRING
//!
////////////////////////////////////////////////////////////////////////////////

template<typename ENUMERATED_TYPE>
void	EnumeratedParameterDescriptor<ENUMERATED_TYPE>::Assign(const char*
		String, Bitmap::CorrectionParameters& Parameters) const
		noexcept(false)
	{
	register	unsigned int	Index;

	if (*String != EQUAL_CHARACTER)
		throw EnumeratedException(ExceptionKind);
	++String;
	for (Index = 0; Index < GetNumberOfValues(); ++Index)
		if (strcmp(String, ValueString [Index]) == 0)
			{
			if (String [strlen(ValueString [Index])] != END_STRING)
				throw EnumeratedException(ExceptionKind);
			Parameters.*ParameterPointer =
				static_cast<ENUMERATED_TYPE>(Index);
			return;
			}
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Write on standard output a string describing a command line parameter of
//!	an enumerated type.
//!
//! \param	DefaultParameters		Reference to the
//!						CorrectionParameter struct which
//!						contains the default parameters.
//!
//! \throw	FileExceptionType::FILE_WRITE_ERROR
//!						If an error occurs while writing
//!						the string.
//!
//! \returns	None.
//!
//! \see	STANDARD_OUTPUT
//! \see	EQUAL_CHARACTER
//! \see	SEPARATOR_CHARACTER
//!
////////////////////////////////////////////////////////////////////////////////

template<typename ENUMERATED_TYPE>
void	EnumeratedParameterDescriptor<ENUMERATED_TYPE>::PrintUsage(const
		Bitmap::CorrectionParameters& DefaultParameters) const
		noexcept(false)
	{
	register	unsigned int	Index;

	if (printf(UsageFormatHead, GetName(), EQUAL_CHARACTER) < 0)
		throw FileException(FileExceptionType::FILE_WRITE_ERROR,
			STANDARD_OUTPUT);
	if (fputs(ValueString [0], stdout) < 0)
		throw FileException(FileExceptionType::FILE_WRITE_ERROR,
			STANDARD_OUTPUT);
	for (Index = 1; Index < GetNumberOfValues(); ++Index)
		{
		if (fputc(SEPARATOR_CHARACTER, stdout) < 0)
			throw FileException(FileExceptionType::FILE_WRITE_ERROR,
				STANDARD_OUTPUT);
		if (fputs(ValueString [Index], stdout) < 0)
			throw FileException(FileExceptionType::FILE_WRITE_ERROR,
				STANDARD_OUTPUT);
		}
	if (printf(UsageFormatTail,
		ValueString [static_cast<unsigned int>(DefaultParameters.*
		ParameterPointer)]) < 0)
		throw FileException(FileExceptionType::FILE_WRITE_ERROR,
			STANDARD_OUTPUT);
	}

}
