////////////////////////////////////////////////////////////////////////////////
//! \file
//!
//! \author	Mauro Fiorentini
//!
//! \date	21/09/2021
//!
//! \par	Function:
//!	Contains the declaration of the GraphicSpace::IntegerException class.
//!
////////////////////////////////////////////////////////////////////////////////

# if	!defined INTEGER_EXCEPTION_H

# define	INTEGER_EXCEPTION_H

# include	"Exception.h"

namespace	GraphicSpace
{

////////////////////////////////////////////////////////////////////////////////
//! \enum	IntegerExceptionType
//!	The IntegerExceptionType type represents the exceptions thrown when an
//!	integer value is outside the valid range.
//!
//! \note
//!	The sequence of values must correspond to the Message array in the
//!	IntegerException class.
//!
//! \see	IntegerException
//!
////////////////////////////////////////////////////////////////////////////////

enum	class	IntegerExceptionType
	{

		//! Invalid file size.
	INVALID_FILE_SIZE,

		//! Invalid bitmap offset.
	INVALID_BITMAP_OFFSET,

		//! Invalid DIB header size.
	INVALID_DIB_HEADER_SIZE,

		//! Invalid width of the image.
	INVALID_IMAGE_WIDTH,

		//! Invalid height of the image.
	INVALID_IMAGE_HEIGHT,

		//! Bitmap too large.
	BITMAP_TOO_LARGE,

		//! Invalid number of color planes.
	INVALID_COLOR_PLANES,

		//! Invalid number of bits per pixel.
	INVALID_BITS_PER_PIXEL,

		//! Invalid value for compression.
	INVALID_COMPRESSION,

		//! Invalid image size.
	INVALID_IMAGE_SIZE,

		//! Invalid number of colors.
	INVALID_NUMBER_OF_COLORS,

		//! Invalid number of important colors.
	INVALID_NUMBER_OF_IMPORTANT_COLORS,

		//! Invalid number of threads parameter.
	INVALID_NUMBER_OF_THREADS,

		//! Invalid number of termites parameter.
	INVALID_NUMBER_OF_TERMITES,

		//! Invalid number of steps parameter.
	INVALID_NUMBER_OF_STEPS

	};

////////////////////////////////////////////////////////////////////////////////
//! \class	IntegerException
//!	The IntegerException class represents an exceptions thrown because an
//!	integer number is outside its valid range.
//!
////////////////////////////////////////////////////////////////////////////////

class	IntegerException: private Exception
	{
	private:

//! \property	ExceptionKind
//!	Kind of the exception.
			IntegerExceptionType	ExceptionKind;

//! \property	Value
//!	Value that caused the exception.
			uint_fast32_t		Value;

//! \property	ErrorFormat
//!	Format string used to write error messages with a number.
		static	const	char*	const	ErrorFormat;

//! \property	Message
//!	Array of error messages, one for each error.
//!
//! \note
//!	The sequence of messages must correspond to the sequence of values of
//!	the IntegerExceptionType type.
//!
//! \see IntegerExceptionType 
		static	const	char*	const	Message [];

	public:

			IntegerException(IntegerExceptionType thisExceptionKind,
				uint_fast32_t thisValue) noexcept;

			~IntegerException(void) noexcept = default;

		void	Write(FILE* OutputFile) const noexcept override;

	};

}

# endif
