////////////////////////////////////////////////////////////////////////////////
//! \file
//!
//! \author	Mauro Fiorentini
//!
//! \date	19/11/2021
//!
//! \par	Function:
//!	Contains the definitions of inline functions of the
//!	GraphicSpace::Interpolator class.
//!
////////////////////////////////////////////////////////////////////////////////

namespace	GraphicSpace
{

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Linearly interpolate a function inside a square.
//!
//! \param	Row		Row index.
//! \param	Column		Column index.
//! \param	UpperLeft	Value of the function in the upper left corner.
//! \param	UpperRight	Value of the function in the upper right corner.
//! \param	LowerLeft	Value of the function in the lower left corner.
//! \param	LowerRight	Value of the function in the lower right corner.
//!
//! \throw	None.
//!
//! \returns	uint_least_32_t	The interpolated value.
//!
//! \note
//!	If Row or column are not less than size, the behavior is undefined.
//!
//! \note
//!	If UpperLeft, UpperRight, LowerLeft or LowerRight are not less than
//!	UINT_LEAST32_MAX / (Size * Size), the returned value is undefined.
//!
////////////////////////////////////////////////////////////////////////////////

inline	uint_least32_t	Interpolator::operator () (unsigned int Row,
				unsigned int Column, uint_least32_t UpperLeft,
				uint_least32_t UpperRight,
				uint_least32_t LowerLeft,
				uint_least32_t LowerRight) const noexcept
	{
	assert(Row < Size);
	assert(Column < Size);
	return (UpperLeft * InterpolationTable [Row] [Column].UpperLeft +
		UpperRight * InterpolationTable [Row] [Column].UpperRight +
		LowerLeft * InterpolationTable [Row] [Column].LowerLeft +
		LowerRight * InterpolationTable [Row] [Column].LowerRight) /
		SquaredSize;
	}

}
