////////////////////////////////////////////////////////////////////////////////
//! \file
//!
//! \author	Mauro Fiorentini
//!
//! \date	17/10/2021
//!
//! \par	Function:
//!	Contains the declaration of the GraphicSpace::NumericParameterDescriptor
//!	class template.
//!
////////////////////////////////////////////////////////////////////////////////

# if	!defined NUMERIC_PARAMETER_DESCRIPTOR_H

# define	NUMERIC_PARAMETER_DESCRIPTOR_H

# include	"Bitmap.h"
# include	"ParameterDescriptor.h"
# include	"FileException.h"

namespace	GraphicSpace
{

////////////////////////////////////////////////////////////////////////////////
//! \class	NumericParameterDescriptor
//!	The NumericParameterDescriptor class template represents the descriptor
//!	of a command line parameter with integer value.
//!
//! \tparam	NUMERIC_TYPE	The type of the parameter; it must be either
//!				uint_fast32_t or float.
//! \tparam	EXCEPTION_TYPE	The type of the exception generated if a string
//!				to be converted is not composed only of digits
//!				or if it represents a number less than Minimum
//!				or greater than Maximum. It must be a predefined
//!				or enumerated type.
//!
////////////////////////////////////////////////////////////////////////////////

template<typename NUMERIC_TYPE, typename EXCEPTION_TYPE>
class	NumericParameterDescriptor: public ParameterDescriptor
	{

	private:

//! \property	Minimum
//!	Minimum valid value of the parameter.
		NUMERIC_TYPE	Minimum;

//! \property	Maximum
//!	Maximum valid value of the parameter.
		NUMERIC_TYPE	Maximum;

//! \property	ParameterPointer
//!	Member pointer to the parameter within the CorrectionParameter struct.
		NUMERIC_TYPE	Bitmap::CorrectionParameters::*
					ParameterPointer;

//! \property	ExceptionKind
//!	The kind of exception that is generated if a string to be converted is
//!	not composed only of digits or if it represents a number less than
//!	Minimum or greater than Maximum.
		EXCEPTION_TYPE	ExceptionKind;

//! \property	UsageFormat
//!	Format string used to write the usage message related to a numeric
//!	parameter.
		static	const	char*	UsageFormat;

				NumericParameterDescriptor(const
					NumericParameterDescriptor& Descriptor)
					noexcept = delete;

				NumericParameterDescriptor
					(NumericParameterDescriptor&&
					Descriptor) noexcept = delete;

		void		operator = (const NumericParameterDescriptor&
					Descriptor) noexcept = delete;

		void		operator = (NumericParameterDescriptor&&
					Descriptor) noexcept = delete;

	public:

				NumericParameterDescriptor(const char* thisName,
					NUMERIC_TYPE thisMinimum,
					NUMERIC_TYPE thisMaximum, NUMERIC_TYPE
					Bitmap::CorrectionParameters::*
					thisParameterPointer,
					EXCEPTION_TYPE thisException) noexcept;

		virtual		~NumericParameterDescriptor(void) noexcept =
					default;

		NUMERIC_TYPE	GetMinimum(void) const noexcept;

		NUMERIC_TYPE	GetMaximum(void) const noexcept;

		void		Assign(const char* String,
					Bitmap::CorrectionParameters&
					Parameters) const noexcept(false)
					override;

		void		PrintUsage(const Bitmap::CorrectionParameters&
					DefaultParameters) const
					noexcept(false) override;

	};

}

# include	"NumericParameterDescriptor.tpl"

# endif
