////////////////////////////////////////////////////////////////////////////////
//! \file
//!
//! \author	Mauro Fiorentini
//!
//! \date	17/10/2021
//!
//! \par	Function:
//!	It contains the definitions of the functions of the
//!	GraphicSpace::NumericParameterDescriptor class template.
//!
////////////////////////////////////////////////////////////////////////////////

namespace	GraphicSpace
{

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Constructor.
//!
//! \param	thisName		Pointer to the string containing the
//!					name of the parameter.
//! \param	thisMinimum		Minimum value for the parameter.
//! \param	thisMaximum		Maximum value for the parameter.
//! \param	thisParameterPointer	Member pointer to the parameter within
//!					the CorrectionParameter struct.
//! \param	thisException		The kind of exception that is generated
//!					if a string to be converted is not
//!					composed only of digits or if it
//!					represents a number less than Minimum or
//!					greater than Maximum.
//!
//! \throw	None.
//!
//! \returns	Not applicable.
//!
//! \note
//!	if thisName is not a valid pointer to a properly terminated string, the
//!	behavior is undefined.
//!
////////////////////////////////////////////////////////////////////////////////

template<typename NUMERIC_TYPE, typename EXCEPTION_TYPE>
inline	NumericParameterDescriptor<NUMERIC_TYPE, EXCEPTION_TYPE>::
		NumericParameterDescriptor(const char* thisName,
		NUMERIC_TYPE thisMinimum, NUMERIC_TYPE thisMaximum,
		NUMERIC_TYPE Bitmap::CorrectionParameters::*
		thisParameterPointer, EXCEPTION_TYPE thisException) noexcept:
		ParameterDescriptor(thisName), Minimum(thisMinimum),
		Maximum(thisMaximum), ParameterPointer(thisParameterPointer),
		ExceptionKind(thisException)
	{
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Get the minimum value of the parameter.
//!
//! \throw	None.
//!
//! \returns	NUMERIC_TYPE	The minimum value of the parameter.
//!
////////////////////////////////////////////////////////////////////////////////

template<typename NUMERIC_TYPE, typename EXCEPTION_TYPE>
inline	NUMERIC_TYPE	NumericParameterDescriptor<NUMERIC_TYPE,
				EXCEPTION_TYPE>::GetMinimum(void) const noexcept
	{
	return Minimum;
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Get the maximum value of the parameter.
//!
//! \throw	None.
//!
//! \returns	NUMERIC_TYPE	The maximum value of the parameter.
//!
////////////////////////////////////////////////////////////////////////////////

template<typename NUMERIC_TYPE, typename EXCEPTION_TYPE>
inline	NUMERIC_TYPE	NumericParameterDescriptor<NUMERIC_TYPE,
				EXCEPTION_TYPE>::GetMaximum(void) const noexcept
	{
	return Maximum;
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Write on standard output a string describing a numeric command line
//!	parameter.
//!
//! \param	DefaultParameters		Reference to the
//!						CorrectionParameter struct which
//!						contains the default parameters.
//!
//! \throw	FileExceptionType::FILE_WRITE_ERROR
//!						If an error occurs while writing
//!						the string.
//!
//! \returns	None.
//!
//! \see	STANDARD_OUTPUT
//! \see	EQUAL_CHARACTER
//!
////////////////////////////////////////////////////////////////////////////////

template<typename NUMERIC_TYPE, typename EXCEPTION_TYPE>
inline	void	NumericParameterDescriptor<NUMERIC_TYPE, EXCEPTION_TYPE>::
			PrintUsage(const Bitmap::CorrectionParameters&
			DefaultParameters) const noexcept(false)
	{
	if (printf(UsageFormat, GetName(), EQUAL_CHARACTER, Minimum, Maximum,
		DefaultParameters.*ParameterPointer) < 0)
		throw FileException(FileExceptionType::FILE_WRITE_ERROR,
			STANDARD_OUTPUT);
	}

}
