////////////////////////////////////////////////////////////////////////////////
//! \file
//!
//! \author	Mauro Fiorentini
//!
//! \date	27/09/2021
//!
//! \par	Function:
//!	It contains the definitions of inline functions of the
//!	GraphicSpace::RandomGenerator class.
//!
////////////////////////////////////////////////////////////////////////////////

namespace	GraphicSpace
{

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Constructor.
//!
//! \throw	None.
//!
//! \returns	Not applicable.
//!
//! \see	NUMBER_OF_GENERATORS
//!
////////////////////////////////////////////////////////////////////////////////

inline	RandomGenerator::RandomGenerator(void) noexcept:
		Selector(229, 17, 0x10000),
		Generator
			{
			SimpleRandomGenerator(65535 + 1, 293, 4294836225),
			SimpleRandomGenerator(32767 * 4 + 1, 283, 4294705156),
			SimpleRandomGenerator(923 + 1, 37, 4294574089),
			SimpleRandomGenerator(65531 + 1, 281, 4294311961),
			SimpleRandomGenerator(65521 + 1, 277, 4293001441),
			SimpleRandomGenerator(65519 + 1, 271, 4292739361),
			SimpleRandomGenerator(65509 + 1, 269, 4291429081),
			SimpleRandomGenerator(65497 + 1, 269, 4289857009),
			SimpleRandomGenerator(65491 + 1, 257, 4289071081),
			SimpleRandomGenerator(65489 + 1, 251, 4288809121),
			SimpleRandomGenerator(65483 + 1, 241, 4288023289),
			SimpleRandomGenerator(65479 + 1, 239, 4287499441),
			SimpleRandomGenerator(65477 + 1, 233, 4287237529),
			SimpleRandomGenerator(65473 + 1, 229, 4286713729),
			SimpleRandomGenerator(65459 + 1, 227, 4284880681),
			SimpleRandomGenerator(65453 + 1, 223, 4284095209),
			SimpleRandomGenerator(65449 + 1, 211, 4283571601),
			SimpleRandomGenerator(65447 + 1, 199, 4283309809),
			SimpleRandomGenerator(65437 + 1, 197, 4282000969),
			SimpleRandomGenerator(65431 + 1, 193, 4281215761),
			SimpleRandomGenerator(65423 + 1, 191, 4280168929),
			SimpleRandomGenerator(65419 + 1, 181, 4280168929),
			SimpleRandomGenerator(65413 + 1, 179, 4278860569),
			SimpleRandomGenerator(65411 + 1, 173, 4278598921),
			SimpleRandomGenerator(65407 + 1, 167, 4278075649),
			SimpleRandomGenerator(65393 + 1, 163, 4276244449),
			SimpleRandomGenerator(65389 + 1, 157, 4275721321),
			SimpleRandomGenerator(65381 + 1, 151, 4274675161),
			SimpleRandomGenerator(65371 + 1, 149, 4273367641)
			}
	{
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Set the seeds of the generator.
//!
//! \param	Seed	Initial value.
//!
//! \throw	None.
//!
//! \returns	None.
//!
//! \see	NUMBER_OF_GENERATORS
//!
////////////////////////////////////////////////////////////////////////////////

inline	void	RandomGenerator::SetSeed(const uint_fast32_t Seed
			[NUMBER_OF_GENERATORS + 1]) noexcept
	{
	register	unsigned int	Index;

	for (Index = 0; Index < NUMBER_OF_GENERATORS; ++Index)
		Generator [Index].SetSeed(Seed [Index + 1]);
	Selector.SetSeed(Seed [NUMBER_OF_GENERATORS]);
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Get the seeds of the generator.
//!
//! \param	Seed	Array that will contain the seeds.
//!
//! \throw	None.
//!
//! \returns	None.
//!
//! \see	NUMBER_OF_GENERATORS
//!
////////////////////////////////////////////////////////////////////////////////

inline	void	RandomGenerator::GetSeed(uint_fast32_t Seed
			[NUMBER_OF_GENERATORS + 1]) const noexcept
	{
	register	unsigned int	Index;

	for (Index = 0; Index < NUMBER_OF_GENERATORS; ++Index)
		Seed [Index + 1] = Generator [Index].GetValue();
	Seed [NUMBER_OF_GENERATORS] = Selector.GetValue();
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Modify the seeds of the generator, xor-ing them with a given value.
//!
//! \param	Value	The value to be xor-ed with the seeds.
//!
//! \throw	None.
//!
//! \returns	None.
//!
//! \see	NUMBER_OF_GENERATORS
//!
////////////////////////////////////////////////////////////////////////////////

inline	void	RandomGenerator::ModifySeed(uint_fast32_t Value) noexcept
	{
	register	unsigned int	Index;

	for (Index = 0; Index < NUMBER_OF_GENERATORS; ++Index)
		Generator [Index].SetSeed(Generator [Index].GetValue() ^ Value);
	Selector.SetSeed(Selector.GetValue() ^ Value);
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Generate a new random number, between 0 and Maximum, with uniform
//!	probability.
//!
//! \param	Maximum		The maximum number that can be generated; if it
//!				is zero, zero is generated;
//!
//! \throw	None.
//!
//! \returns	uint_fast32_t	The generated number.
//!
//! \note
//!	A new value is computed selecting generator with the Selector generator
//!	and then using it to generate the number.
//!
//! \note
//!	If Maximum is not less than MINIMUM_MODULE, the behavior is undefined.
//!
//! \see	NUMBER_OF_GENERATORS
//!
////////////////////////////////////////////////////////////////////////////////

inline	uint_fast32_t	RandomGenerator::Generate(uint_fast32_t Maximum)
				noexcept
	{
	return Generator [Selector.Generate(NUMBER_OF_GENERATORS - 1)].
		Generate(Maximum);
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Select among some alternatives, with a given probability distribution.
//!
//! \param	Probability	Pointer to the array of probabilities, which
//!				specify the probability distribution.
//! \param	Elements	Number of emenents among which the choice is
//!				made.
//!
//! \throw	None.
//!
//! \returns	unsigned int	The selected item; the value is in the range
//!				[0 .. Elements - 1], or UINT_MAX, if all
//!				elements of the Probability array are zero.
//!
//! \note
//!	If Pointer is not a pointer to an array of at least Elements values, the
//!	behavior is undefined.
//!
//! \note
//!	If the sum of all the probabilities is not less than MINIMUM_MODULE, the
//!	behavior is undefined.
//!
//! \see	UINT_MAX
//! \see	UINT_MAX
//!
////////////////////////////////////////////////////////////////////////////////

inline	unsigned int	RandomGenerator::Select(register const uint_fast32_t*
				Probability, unsigned int Elements) noexcept
	{
	register		uint_fast32_t	Value;
	register	const	uint_fast32_t*	Pointer;

	Value = 0;
	Pointer = Probability + Elements;
	while (Probability < Pointer)
		Value += *Probability++;
	if (Value == 0)
		return UINT_MAX;
	Value = Generate(Value - 1);
	Pointer = Probability -= Elements;
	for (;;)
		{
		if (Value < *Probability)
			return static_cast<unsigned int>(Probability - Pointer);
		Value -= *Probability++;
		}
	}

}
