////////////////////////////////////////////////////////////////////////////////
//! \file
//!
//! \author	Mauro Fiorentini
//!
//! \date	23/09/2021
//!
//! \par	Function:
//!	It contains the declaration of the GraphicSpace::SimpleRandomGenerator
//!	class.
//!
////////////////////////////////////////////////////////////////////////////////

# if !defined	SIMPLE_RANDOM_GENERATOR_H

# define	SIMPLE_RANDOM_GENERATOR_H

namespace	GraphicSpace
{

////////////////////////////////////////////////////////////////////////////////
//! \class	SimpleRandomGenerator
//!	The SimpleRandomGenerator class represents a linear congruent random
//!	numbers generator. \n
//!	Proper choice of the parameter ensures that the cycle of generated
//!	numbers repeats itself only after Module generations. \n
//!	See Erwin Donald Knuth, The art of Computer Programming, vol II.
//!
////////////////////////////////////////////////////////////////////////////////

class	SimpleRandomGenerator
	{
	private:

//! \property	Multiplier
//!	Multiplier used by the random number generator.
//!
//! \note
//!	It must be a prime number less than UINT32_MAX and of the form
//!	4 * n + 1.
					uint_fast32_t	Multiplier;

//! \property	Constant
//!	Additive constant used by the random number generator.
					uint_fast32_t	Constant;

//! \property	Module
//!	Module used by the random number generator.
					uint_fast32_t	Module;

//! \property	Value
//!	Latest generated value.
					uint_fast32_t	Value;

//! \property	DEFAULT_MULTIPLIER
//!	Default multiplier value.
		static	constexpr	uint_fast32_t	DEFAULT_MULTIPLIER = 0;

//! \property	DEFAULT_CONSTANT
//!	Default constant value.
		static	constexpr	uint_fast32_t	DEFAULT_CONSTANT = 0;

//! \property	DEFAULT_MODULE
//!	Default module value.
		static	constexpr	uint_fast32_t	DEFAULT_MODULE = 0;

//! \property	DEFAULT_SEED
//!	Default seed value.
		static	constexpr	uint_fast32_t	DEFAULT_SEED = 0;

		uint_fast32_t		Generate(void) noexcept;

	public:

					SimpleRandomGenerator(void) noexcept;

					SimpleRandomGenerator(const
						SimpleRandomGenerator&
						Argument) noexcept = default;

					SimpleRandomGenerator(
						SimpleRandomGenerator&&
						Argument) noexcept = default;

					SimpleRandomGenerator(uint_fast32_t
						thisMultiplier,
						uint_fast32_t thisConstant,
						uint_fast32_t thisModule)
						noexcept;

		SimpleRandomGenerator&	operator = (const SimpleRandomGenerator&
						Argument) noexcept = default;

		SimpleRandomGenerator&	operator = (SimpleRandomGenerator&&
						Argument) noexcept = default;

					~SimpleRandomGenerator(void) noexcept =
						default;

		void			SetSeed(uint_fast32_t thisConstant)
						noexcept;

		uint_fast32_t		GetValue(void) const noexcept;

		uint_fast32_t		Generate(uint_fast32_t Maximum)
						noexcept;

	};

}

# include	"SimpleRandomGenerator.inl"

# endif
