////////////////////////////////////////////////////////////////////////////////
//! \file
//!
//! \author	Mauro Fiorentini
//!
//! \date	23/09/2021
//!
//! \par	Function:
//!	It contains the definitions of inline functions of the
//!	GraphicSpace::SimpleRandomGenerator class.
//!
////////////////////////////////////////////////////////////////////////////////

namespace	GraphicSpace
{

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Generate a new random number, between 0 and Module, with uniform
//!	probability.
//!
//! \throw	None.
//!
//! \returns	uint_fast32_t	the generated value.
//!
//! \note
//!	A new value is computed as (Value * Multiplier + Constant) % Module and
//!	and it is stored in Value.
//!
////////////////////////////////////////////////////////////////////////////////

inline	uint_fast32_t	SimpleRandomGenerator::Generate(void) noexcept
	{
	return Value = (static_cast<uint_fast64_t>(Value) *
		static_cast<uint_fast64_t>(Multiplier) +
		static_cast<uint_fast64_t>(Constant)) % Module;
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Constructor; initialize the object with default values.
//!
//! \throw	None.
//!
//! \returns	Not applicable.
//!
//! \see	DEFAULT_MULTIPLIER
//! \see	DEFAULT_CONSTANT
//! \see	DEFAULT_MODULE
//! \see	DEFAULT_SEED
//!
////////////////////////////////////////////////////////////////////////////////

inline	SimpleRandomGenerator::SimpleRandomGenerator(void) noexcept:
		Multiplier(DEFAULT_MULTIPLIER), Constant(DEFAULT_CONSTANT),
		Module(DEFAULT_MODULE), Value(DEFAULT_SEED)
	{
	assert(Module < UINT32_MAX);
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Constructor; initialize Value with DEFAULT_SEED.
//!
//! \param	thisMultiplier	Multiplier.
//! \param	thisConstant	Constant.
//! \param	thisModule	Module.
//!
//! \throw	None.
//!
//! \returns	Not applicable.
//!
//! \note
//!	If Module is not less than UINT32_MAX, the behavior is undefined.
//!
//! \see	DEFAULT_SEED
//!
////////////////////////////////////////////////////////////////////////////////

inline	SimpleRandomGenerator::SimpleRandomGenerator(uint_fast32_t
		thisMultiplier,  uint_fast32_t thisConstant,
		uint_fast32_t thisModule) noexcept: Multiplier(thisMultiplier),
		Constant(thisConstant), Module(thisModule), Value(DEFAULT_SEED)
	{
	assert(Module < UINT32_MAX);
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Set the seed of the generator.
//!
//! \param	Seed	Initial value.
//!
//! \throw	None.
//!
//! \returns	Not applicable.
//!
////////////////////////////////////////////////////////////////////////////////

inline	void	SimpleRandomGenerator::SetSeed(uint_fast32_t Seed) noexcept
	{
	Value = Seed % Module;
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Get current seed value, equal to latest generated number.
//!
//! \throw	None.
//!
//! \returns	uint_fast32_t	The seed value.
//!
////////////////////////////////////////////////////////////////////////////////

inline	uint_fast32_t	SimpleRandomGenerator::GetValue(void) const noexcept
	{
	return Value;
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Generate a new random number, between 0 and Maximum, with uniform
//!	probability.
//!
//! \param	Maximum		The maximum number that can be generated; if it
//!				is zero, zero is generated;
//!
//! \throw	None.
//!
//! \returns	uint_fast32_t	The generated number.
//!
//! \note
//!	A new value is computed generating a random number and then verifying if
//!	it is less than Module - (Module + 1) % (Maximum + 1); if it is not
//!	less, it is discarded and a new random number is generated. This ensures
//!	that the generated numbers have a uniform probability distribution.
//!
//! \note
//!	The algorithm used to generate the random numbers guarantees that a
//!	valid number is generated quickly: less that two attempts are necessary
//!	on average.
//!
//! \note
//!	If Maximum is not less than Module, the behavior is undefined.
//!
////////////////////////////////////////////////////////////////////////////////

inline	uint_fast32_t	SimpleRandomGenerator::Generate(uint_fast32_t Maximum)
				noexcept
	{
	register	uint_fast32_t	MaximumValue;

	assert(Maximum < Module);
	if (Maximum == 0)
		return 0;
	MaximumValue = Module - (Module + 1) % (Maximum + 1);
	while (Generate() > MaximumValue);
	return Value % (Maximum + 1);
	}

}
