////////////////////////////////////////////////////////////////////////////////
//! \file
//!
//! \author	Mauro Fiorentini
//!
//! \date	15/08/2022
//!
//! \par	Function:
//!	Contains the declaration of the GraphicSpace::SystemException class.
//!
////////////////////////////////////////////////////////////////////////////////

# if	!defined SYSTEM_EXCEPTION_H

# define	SYSTEM_EXCEPTION_H

# include	"Exception.h"

namespace	GraphicSpace
{

////////////////////////////////////////////////////////////////////////////////
//! \enum	SystemExceptionType
//!	The SystemExceptionType type represents the system exceptions.
//!
//! \note
//!	The sequence of values must correspond to the Message array in the
//!	SystemException class.
//!
//! \see	SystemException
//!
////////////////////////////////////////////////////////////////////////////////

enum	class	SystemExceptionType
	{
		// Error while initializing the thread management.
	THREAD_INITIALIZATION_FAILED,

		// Thread creation failure.
	THREAD_CREATION_FAILED,

		// Invalid operation on a thread.
	INVALID_THREAD_OPERATION,

		// Thread join failure.
	THREAD_JOIN_FAILED,

		// Error while initializing a mutex.
	MUTEX_INITIALIZATION_FAILED,

		// Invalid operation on a mutex.
	INVALID_MUTEX_OPERATION,

		// Error while destroying a mutex.
	MUTEX_DESTROY_FAILED,

		// Mutex lock failure.
	MUTEX_LOCK_FAILED,

		// Mutex unlock failure.
	MUTEX_UNLOCK_FAILED

	};

////////////////////////////////////////////////////////////////////////////////
//! \class	SystemException
//!	The SystemException class represents an exceptions throw during a thread
//!	management operation.
//!
////////////////////////////////////////////////////////////////////////////////

class	SystemException: private Exception
	{
	private:

//! \property	ExceptionKind
//!	Kind of the exception.
			SystemExceptionType	ExceptionKind;

//! \property	ErrorFormat
//!	Format string used to write error messages.
		static	const	char*	const	ErrorFormat;

//! \property	Message
//!	Array of error messages, one for each error.
//!
//! \note
//!	The sequence of messages must correspond to the sequence of values of
//!	the SystemExceptionType type.
//!
//! \see SystemExceptionType 
		static	const	char*	const	Message [];

	public:

				SystemException(SystemExceptionType
					thisExceptionKind) noexcept;

				SystemException(const SystemException& Argument)
					noexcept = default;

				SystemException(SystemException&& Argument)
					noexcept = default;

				~SystemException(void) noexcept = default;

		SystemException&	operator = (const SystemException&
						Argument) noexcept = default;

		SystemException&	operator = (SystemException&& Argument)
						noexcept = default;

		void		Write(FILE* OutputFile) const noexcept override;

	};

}

# endif
