////////////////////////////////////////////////////////////////////////////////
//! \file
//!
//! \author	Mauro Fiorentini
//!
//! \date	12/09/2021
//!
//! \par	Function:
//!	Contains the definitions of non inline functions of the
//!	GraphicSpace::Bitmap class.
//!
////////////////////////////////////////////////////////////////////////////////

# include	"Graphic.h"
# include	"Bitmap.h"
# include	"FileException.h"
# include	"IntegerException.h"
# include	"FloatException.h"
# include	"NameException.h"
# include	"Interpolator.h"
# include	"Thread.h"

# include	<stdio.h>
# include	<string.h>
# include	<math.h>

constexpr	unsigned int	Step = 5;

namespace	GraphicSpace
{

Interpolator	BitmapInterpolator(Step);

const	char		Bitmap::HeaderName [HEADER_NAME_LENGTH] = { 'B', 'M' };

const	char*	const	Bitmap::OpenReadBinary = "rb";

const	char*	const	Bitmap::OpenWriteBinary = "wb";

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Compute the index of next pixel to visit, choosing the one with the
//!	maximum probability. \n
//!	If all neighbours have probability zero, the function returns
//!	UINT_FAST_32_MAX:
//!
//! \param	CurrentGenerator	Reference to the random number
//!					generator.
//! \param	RepulsionPointer	Array of the repulsion value for each
//!					pixel.
//!
//! \throw	None.
//!
//! \returns	IndexType		The index of next pixel.
//!
//! \note
//!	If RepulsionPointer is not a pointer to an array of at least BitmapSize
//!	elements, the behavior is undefined.
//!
//! \note
//!	The returned value is the index of the pixel with the maximum
//!	probability among neighbours; if two or more pixels have the same
//!	probability, the random number generator is used to randomly choose one.
//!
//! \note
//!	A bit of loop unrolling helps generating faster code.
//!
//! \see	NUMBER_OF_NEIGHBOURS
//! \see	UINT_FAST_32_MAX
//!
////////////////////////////////////////////////////////////////////////////////

Bitmap::IndexType	Bitmap::PixelInformationType::
				NextPixelMaximum(RandomGenerator&
				CurrentGenerator,
				register const RepulsionType* RepulsionPointer)
				const noexcept
	{
	register	unsigned int	Best;
	register	uint_fast32_t	Value;
	register	uint_fast32_t	Maximum;
			IndexType	BestIndex [NUMBER_OF_NEIGHBOURS];

	assert(NumberOfNeighbours == 3 or NumberOfNeighbours == 5 or
		NumberOfNeighbours == 8);
	Best = 1;
	Maximum = Neighbour [0].Attraction *
		RepulsionPointer [Neighbour [0].Index].Repulsion;
	BestIndex [0] = Neighbour [0].Index;
	if ((Value = Neighbour [1].Attraction *
		RepulsionPointer [Neighbour [1].Index].Repulsion) >= Maximum)
		if (Value > Maximum)
			{
			Maximum = Value;
			BestIndex [0] = Neighbour [1].Index;
			}
		else
			{
			BestIndex [1] = Neighbour [1].Index;
			Best = 2;
			}
	if ((Value = Neighbour [2].Attraction *
		RepulsionPointer [Neighbour [2].Index].Repulsion) >= Maximum)
		if (Value > Maximum)
			{
			Maximum = Value;
			BestIndex [0] = Neighbour [2].Index;
			Best = 1;
			}
		else
			BestIndex [Best++] = Neighbour [2].Index;
	if (NumberOfNeighbours > 3)
		{
		if ((Value = Neighbour [3].Attraction *
			RepulsionPointer [Neighbour [3].Index].Repulsion) >=
			Maximum)
			if (Value > Maximum)
				{
				Maximum = Value;
				BestIndex [0] = Neighbour [3].Index;
				Best = 1;
				}
			else
				BestIndex [Best++] = Neighbour [3].Index;
		if ((Value = Neighbour [4].Attraction *
			RepulsionPointer [Neighbour [4].Index].Repulsion) >=
			Maximum)
			if (Value > Maximum)
				{
				Maximum = Value;
				BestIndex [0] = Neighbour [4].Index;
				Best = 1;
				}
			else
				BestIndex [Best++] = Neighbour [4].Index;
		if (NumberOfNeighbours > 5)
			{
			if ((Value = Neighbour [5].Attraction *
				RepulsionPointer [Neighbour [5].Index].
				Repulsion) >= Maximum)
				if (Value > Maximum)
					{
					Maximum = Value;
					BestIndex [0] = Neighbour [5].Index;
					Best = 1;
					}
				else
					BestIndex [Best++] =
						Neighbour [5].Index;
			if ((Value = Neighbour [6].Attraction *
				RepulsionPointer [Neighbour [6].Index].
				Repulsion) >= Maximum)
				if (Value > Maximum)
					{
					Maximum = Value;
					BestIndex [0] = Neighbour [6].Index;
					Best = 1;
					}
				else
					BestIndex [Best++] =
						Neighbour [6].Index;
			if ((Value = Neighbour [7].Attraction *
				RepulsionPointer [Neighbour [7].Index].
				Repulsion) >= Maximum)
				if (Value > Maximum)
					{
					Maximum = Value;
					BestIndex [0] = Neighbour [7].Index;
					Best = 1;
					}
				else
					BestIndex [Best++] =
						Neighbour [7].Index;
			}
		}
	if (Best == 1)
		return BestIndex [0];
	if (Maximum == 0)
		return UINT_FAST32_MAX;
	return BestIndex [CurrentGenerator.Generate(Best - 1)];
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Compute the index of next pixel to visit, choosing at random,
//!	accordingly with the probability distribution.
//!	If all neighbours have probability zero, the function returns
//!	UINT_FAST_32_MAX:
//!
//! \param	CurrentGenerator	Reference to the random number
//!					generator.
//! \param	RepulsionPointer	Array of the repulsion value for each
//!					pixel.
//!
//! \throw	None.
//!
//! \returns	IndexType		The index of next pixel.
//!
//! \note
//!	If RepulsionPointer is not a pointer to an array of at least BitmapSize
//!	elements, the behavior is undefined.
//!
//! \note
//!	The returned value is randomly chosen among neighbours, considering the
//!	probability distribution of the pixels.
//!
//! \note
//!	A bit of loop unrolling helps generating faster code.
//!
//! \see	NUMBER_OF_NEIGHBOURS
//! \see	UINT_MAX
//! \see	UINT_FAST_32_MAX
//!
////////////////////////////////////////////////////////////////////////////////

Bitmap::IndexType	Bitmap::PixelInformationType::
				NextPixelRandom(RandomGenerator&
				CurrentGenerator,
				register const RepulsionType* RepulsionPointer)
				const noexcept
	{
	unsigned int	Index;
	uint_fast32_t	Probability [NUMBER_OF_NEIGHBOURS];

	assert(NumberOfNeighbours == 3 or NumberOfNeighbours == 5 or
		NumberOfNeighbours == 8);
	Probability [0] = Neighbour [0].Attraction *
		RepulsionPointer [Neighbour [0].Index].Repulsion;
	Probability [1] = Neighbour [1].Attraction *
		RepulsionPointer [Neighbour [1].Index].Repulsion;
	Probability [2] = Neighbour [2].Attraction *
		RepulsionPointer [Neighbour [2].Index].Repulsion;
	if (NumberOfNeighbours > 3)
		{
		Probability [3] = Neighbour [3].Attraction *
			RepulsionPointer [Neighbour [3].Index].Repulsion;
		Probability [4] = Neighbour [4].Attraction *
			RepulsionPointer [Neighbour [4].Index].Repulsion;
		if (NumberOfNeighbours > 5)
			{
			Probability [5] = Neighbour [5].Attraction *
				RepulsionPointer [Neighbour [5].Index].
				Repulsion;
			Probability [6] = Neighbour [6].Attraction *
				RepulsionPointer [Neighbour [6].Index].
				Repulsion;
			Probability [7] = Neighbour [7].Attraction *
				RepulsionPointer [Neighbour [7].Index].
				Repulsion;
			}
		}
	if ((Index = CurrentGenerator.Select(Probability,
		NumberOfNeighbours)) == UINT_MAX)
		return UINT_LEAST32_MAX;
	return Neighbour [Index].Index;
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Compute the contribution of a termite to the correction of a pixel of a
//!	color.
//!
//! \param	MaximumValue		Maximum value of the color in the whole
//!					image.
//! \param	PixelIndex		Index of the pixel.
//! \param	ColorPointer		Pointer to the array that contains the
//!					the original bitmap of the color.
//! \param	Steps			Number of steps of each termite.
//! \param	PixelInformationPointer	Pointer to the array of information
//!					about the pixels.
//! \param	RepulsionPointer	Array of the repulsion information for
//!					each pixel.
//! \param	CurrentGenerator	Reference to the random number
//!					generator.
//! \param	PathFinder		Member pointer to the function used to
//!					select the pixels to visit.
//!
//! \throw	None.
//!
//! \returns	None.
//!
//! \note
//!	If ColorPointer is not a pointer to an array of at least BitmapSize
//!	elements, the behavior is undefined.
//!
//! \note
//!	If Visitors is not a pointer to an array of at least BitmapSize
//!	elements, the behavior is undefined.
//!
//! \note
//!	If Repulsion is not a pointer to an array of at least BitmapSize
//!	elements, the behavior is undefined.
//!
//! \see	REPULSION_TABLE_SIZE
//!
////////////////////////////////////////////////////////////////////////////////

unsigned int	Bitmap::ComputeTermite(register unsigned int MaximumValue,
			register IndexType PixelIndex,
			register const uint8_t* ColorPointer,
			register unsigned int Steps,
			register const PixelInformationType*
			PixelInformationPointer,
			register RepulsionType* RepulsionPointer,
			register RandomGenerator& CurrentGenerator,
			register PathFunction PathFinder) const noexcept
	{
	register	unsigned int	PathMaximum;

	PathMaximum = ColorPointer [PixelIndex];
	assert(PathMaximum < MaximumValue);
	RepulsionPointer [PixelIndex].Repulsion =
		RepulsionTable [++RepulsionPointer [PixelIndex].Visitors];
	for (; Steps > 0; --Steps)
		{
		if ((PixelIndex = (PixelInformationPointer [PixelIndex].*
			PathFinder)(CurrentGenerator, RepulsionPointer)) >
			BitmapSize)
			break;
		RepulsionPointer [PixelIndex].Repulsion =
			RepulsionTable [++RepulsionPointer [PixelIndex].
			Visitors];
		if (ColorPointer [PixelIndex] > PathMaximum)
			if ((PathMaximum = ColorPointer [PixelIndex]) ==
				MaximumValue)
				break;
		}
	return PathMaximum;
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Correct one color of the bitmap, on an horizontal strip of the bitmap.
//!
//! \param	CorrectedColorPointer	Pointer to the array that will contain
//!					the corrected bitmap of the color.
//! \param	OriginalColorPointer	Pointer to the array that contains the
//!					the original bitmap of the color.
//! \param	StartRow		Index of the first row of the strip.
//! \param	EndRow			Index of the first row after the strip.
//! \param	MaximumValue		Maximum value of the color in the whole
//!					image.
//! \param	Parameters		Reference to the struct containing the
//!					correction parameters.
//! \param	PixelInformationPointer	Pointer to the array of information
//!					about the pixels.
//! \param	PathFinder		Member pointer to the function used to
//!					select the pixels to visit.
//! \param	DivisionTable		Array of quotient: for n > 0 the entry
//!					with index n contains UINT8_MAX *
//!					TERMITES_SUM_SCALE / n, rounded to the
//!					closest integer; the entry with index 0
//!					contains the same value of the entry
//!					with index 1.
//!
//! \throw	None.
//!
//! \returns	None.
//!
//! \note
//!	If CorrectedColorPointer is not a pointer to an array of at least
//!	BitmapSize elements, the behavior is undefined.
//!
//! \note
//!	If OriginalColorPointer is not a pointer to an array of at least
//!	BitmapSize elements, the behavior is undefined.
//!
//! \note
//!	The correction algorithm is applied to all pixel whose row is in the
//!	range [StartRow .. EndRow - 1].
//!
//! \see	UINT8_MAX
//! \see	TERMITES_SUM_SCALE
//!
////////////////////////////////////////////////////////////////////////////////

void	Bitmap::CorrectStrip(register uint8_t* CorrectedColorPointer,
		register const uint8_t* OriginalColorPointer,
		register unsigned int StartRow, register unsigned int EndRow,
		register unsigned int MaximumValue,
		register const CorrectionParameters& Parameters,
		register const PixelInformationType* PixelInformationPointer,
		register PathFunction PathFinder,
		register const uint_least32_t DivisionTable [UINT8_MAX + 1])
		const noexcept
	{
	register	unsigned int	Index;
	register	IndexType	PixelIndex;
	register	IndexType	LastIndex;
	register	uint_fast32_t	Sum;
	register	RepulsionType*	RepulsionPointer;
			RandomGenerator	LocalGenerator(Generator);

	RepulsionPointer = new RepulsionType [BitmapSize];
	LocalGenerator.ModifySeed(StartRow ^ EndRow);
	for (; StartRow < EndRow; ++StartRow)
		{
		LastIndex = MakeIndex(StartRow, Width);
		for (PixelIndex = MakeIndex(StartRow, 0);
			PixelIndex < LastIndex; ++PixelIndex)
			{
			if (OriginalColorPointer [PixelIndex] == MaximumValue)
				continue;
			IgnoreValue(memcpy(RepulsionPointer,
				InitialRepulsionPointer,
				BitmapSize * sizeof(RepulsionType)));
			Sum = 0;
			for (Index = 0; Index < Parameters.Termites; ++Index)
				Sum += DivisionTable
					[ComputeTermite(MaximumValue,
					PixelIndex, OriginalColorPointer,
					Parameters.Steps,
					PixelInformationPointer,
					RepulsionPointer, LocalGenerator,
					PathFinder)];
			CorrectedColorPointer [PixelIndex] =
				static_cast<uint8_t>(RoundedDivide(Sum *
				OriginalColorPointer [PixelIndex],
				TERMITES_SUM_SCALE * Parameters.Termites));
			}
		}
	delete [] RepulsionPointer;
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Correct one color of the bitmap.
//!
//! \param	CorrectedColorPointer	Pointer to the array that will contain
//!					the corrected bitmap of the color.
//! \param	OriginalColorPointer	Pointer to the array that contains the
//!					the original bitmap of the color.
//! \param	MaximumValue		Maximum value of the color in the whole
//!					image.
//! \param	MinimumIndex		Minimimum index of the pixels to
//!					correct.
//! \param	MaximumIndex		Index of the first pixels not to
//!					correct.
//! \param	Parameters		Reference to the struct containing the
//!					correction parameters.
//! \param	PixelInformationPointer	Pointer to the array of information
//!					about the pixels.
//! \param	PathFinder		Member pointer to the function used to
//!					select the pixels to visit.
//! \param	DivisionTable		Array of quotient: for n > 0 the entry
//!					with index n contains UINT8_MAX *
//!					TERMITES_SUM_SCALE / n, rounded to the
//!					closest integer; the entry with index 0
//!					contains the same value of the entry
//!					with index 1.
//!
//! \throw	None.
//!
//! \returns	None.
//!
//! \note
//!	If CorrectedColorPointer is not a pointer to an array of at least
//!	BitmapSize elements, the behavior is undefined.
//!
//! \note
//!	If OriginalColorPointer is not a pointer to an array of at least
//!	BitmapSize elements, the behavior is undefined.
//!
//! \note
//!	The correction algorithm is applied to all pixels of the image.
//!
//! \see	UINT8_MAX
//! \see	TERMITES_SUM_SCALE
//! \see	MAXIMUM_NUMBER_OF_THREADS
//! \see	STRIP_HEIGHT
//!
////////////////////////////////////////////////////////////////////////////////

void	Bitmap::CorrectColor(register uint8_t* CorrectedColorPointer,
		register const uint8_t* OriginalColorPointer,
		register unsigned int MaximumValue,
		register const CorrectionParameters& Parameters,
		register const PixelInformationType* PixelInformationPointer,
		register PathFunction PathFinder,
		register const uint_least32_t DivisionTable [UINT8_MAX + 1])
		noexcept
	{
	register	unsigned int	Index;
	register	unsigned int	StartRow;
			Thread		CorrectorThread
						[MAXIMUM_NUMBER_OF_THREADS];

	if (Parameters.Threads == 1)
		CorrectStrip(CorrectedColorPointer, OriginalColorPointer,
			0, Height, MaximumValue, Parameters,
			PixelInformationPointer, PathFinder, DivisionTable);
	else
		{
		StripMutex.Initialize();
		Strips = (Height + STRIP_HEIGHT - 1) / STRIP_HEIGHT;
		if (DescriptorPointer == nullptr)
			DescriptorPointer =
				new CorrectionDescriptorType [Strips];
		StartRow = 0;
		DescriptorPointer [0].CorrectedColorPointer =
			CorrectedColorPointer;
		DescriptorPointer [0].OriginalColorPointer =
				OriginalColorPointer;
		DescriptorPointer [0].MaximumValue = MaximumValue;
		DescriptorPointer [0].ParametersPointer = &Parameters;
		DescriptorPointer [0].PixelInformationPointer =
			PixelInformationPointer;
		DescriptorPointer [0].PathFinder = PathFinder;
		DescriptorPointer [0].DivisionTablePointer = DivisionTable;
		for (Index = 0; Index < Strips; ++Index)
			{
			DescriptorPointer [Index] = DescriptorPointer [0];
			DescriptorPointer [Index].StartRow = StartRow;
			if ((StartRow += STRIP_HEIGHT) > Height)
				StartRow = Height;
			DescriptorPointer [Index].EndRow = StartRow;
			}

			// Divide the rows as equally as possible among the
			// threads.
		for (Index = 0; Index < Parameters.Threads; ++Index)
			CorrectorThread [Index].Start(CorrectFunction, this);
		for (Index = 0; Index < Parameters.Threads; ++Index)
			CorrectorThread [Index].Join();
		StripMutex.Finalize();
		}
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Return the pointer to next strip to correct, or nullptr, if all strips
//!	have been processed.
//!
//! \throw	None.
//!
//! \returns	const	CorrectionDescriptorType*	The pointer to next
//!							strip to correct.
//!
//! \see	STRIP_INDEX_FORMAT 
//
////////////////////////////////////////////////////////////////////////////////

const	Bitmap::CorrectionDescriptorType*	Bitmap::GetStripDescriptor(void)
							noexcept
	{
	const	Bitmap::CorrectionDescriptorType*	Value;

	StripMutex.Lock();
	if (Strips > 0)
		{
		Value = &DescriptorPointer [--Strips];
		if (not Value->ParametersPointer->Quiet)
			IgnoreValue(printf(STRIP_INDEX_FORMAT, Strips));
		}
	else
		Value = nullptr;
	StripMutex.Unlock();
	return Value;
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Constructor.
//!
//! \throw	None.
//!
//! \returns	Not applicable
//!
//! \see	NUMBER_OF_COLORS
//!
////////////////////////////////////////////////////////////////////////////////

Bitmap::Bitmap(void) noexcept: Width(0), Height(0), BitmapSize(0),
	BitsPerPixel(0), HorizontalResolution(0), VerticalResolution(0),
	PixelPointer(), RepulsionTable(), InitialRepulsionPointer(),
	Generator(), Strips(), StripMutex(), DescriptorPointer(nullptr)
	{
	register	unsigned int	Color;

	for (Color = 0; Color < NUMBER_OF_COLORS; ++Color)
		PixelPointer [Color] = nullptr;
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Copy constructor.
//!
//! \param	thisBitmap	Bitmap to be copied.
//!
//! \throw	None.
//!
//! \returns	Not applicable
//!
//! \see	NUMBER_OF_COLORS
//!
////////////////////////////////////////////////////////////////////////////////

Bitmap::Bitmap(const Bitmap& thisBitmap) noexcept: Width(thisBitmap.Width),
	Height(thisBitmap.Height), BitmapSize(thisBitmap.BitmapSize),
	BitsPerPixel(thisBitmap.BitsPerPixel),
	HorizontalResolution(thisBitmap.HorizontalResolution),
	VerticalResolution(thisBitmap.VerticalResolution), PixelPointer(),
	RepulsionTable(), InitialRepulsionPointer(),
	Generator(thisBitmap.Generator), Strips(thisBitmap.Strips),
	StripMutex(), DescriptorPointer(nullptr)
	{
	register	unsigned int	Color;

	AllocateImage();
	if (BitmapSize != 0)
		for (Color = 0; Color < NUMBER_OF_COLORS; ++Color)
			memcpy(PixelPointer [Color],
				thisBitmap.PixelPointer [Color], BitmapSize);
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Move constructor.
//!
//! \param	thisBitmap	Bitmap to be moved.
//!
//! \throw	None.
//!
//! \returns	Not applicable
//!
//! \see	NUMBER_OF_COLORS
//!
////////////////////////////////////////////////////////////////////////////////

Bitmap::Bitmap(Bitmap&& thisBitmap) noexcept: Width(thisBitmap.Width),
	Height(thisBitmap.Height), BitmapSize(thisBitmap.BitmapSize),
	BitsPerPixel(thisBitmap.BitsPerPixel),
	HorizontalResolution(thisBitmap.HorizontalResolution),
	VerticalResolution(thisBitmap.VerticalResolution), PixelPointer(),
	RepulsionTable(), InitialRepulsionPointer(),
	Generator(thisBitmap.Generator), Strips(thisBitmap.Strips),
	StripMutex(), DescriptorPointer(thisBitmap.DescriptorPointer)
	{
	register	unsigned int	Color;

	for (Color = 0; Color < NUMBER_OF_COLORS; ++Color)
		{
		PixelPointer [Color] = thisBitmap.PixelPointer [Color];
		thisBitmap.PixelPointer [Color] = nullptr;
		}
	thisBitmap.DescriptorPointer = nullptr;
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Destructor.
//!
//! \throw	None.
//!
//! \returns	Not applicable
//!
//! \see	NUMBER_OF_COLORS
//!
////////////////////////////////////////////////////////////////////////////////

Bitmap::~Bitmap(void) noexcept
	{
	DestroyImage();
	if (DescriptorPointer != nullptr)
		delete [] DescriptorPointer;
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Copy assignment.
//!
//! \param	thisBitmap	Bitmap to assign.
//!
//! \throw	None.
//!
//! \returns	Bitmap&		reference to the assigned Bitmap.
//!
//! \see	NUMBER_OF_COLORS
//!
////////////////////////////////////////////////////////////////////////////////

Bitmap&	Bitmap::operator = (const Bitmap& thisBitmap) noexcept
	{
	register	unsigned int	Color;

	Width = thisBitmap.Width;
	Height = thisBitmap.Height;
	BitmapSize = thisBitmap.BitmapSize;
	BitsPerPixel = thisBitmap.BitsPerPixel;
	HorizontalResolution = thisBitmap.HorizontalResolution;
	VerticalResolution = thisBitmap.VerticalResolution;
	Generator = thisBitmap.Generator;
	DestroyImage();
	AllocateImage();
	if (BitmapSize != 0)
		for (Color = 0; Color < NUMBER_OF_COLORS; ++Color)
			memcpy(PixelPointer [Color],
				thisBitmap.PixelPointer [Color], BitmapSize);
	return *this;
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Move assignment.
//!
//! \param	thisBitmap	Bitmap to assign.
//!
//! \throw	None.
//!
//! \returns	Bitmap&		reference to the assigned Bitmap.
//!
//! \see	NUMBER_OF_COLORS
//!
////////////////////////////////////////////////////////////////////////////////

Bitmap&	Bitmap::operator = (Bitmap&& thisBitmap) noexcept
	{
	register	unsigned int	Color;

	Width = thisBitmap.Width;
	Height = thisBitmap.Height;
	BitmapSize = thisBitmap.BitmapSize;
	BitsPerPixel = thisBitmap.BitsPerPixel;
	HorizontalResolution = thisBitmap.HorizontalResolution;
	VerticalResolution = thisBitmap.VerticalResolution;
	Generator = thisBitmap.Generator;
	DestroyImage();
	for (Color = 0; Color < NUMBER_OF_COLORS; ++Color)
		{
		PixelPointer [Color] = thisBitmap.PixelPointer [Color];
		thisBitmap.PixelPointer [Color] = nullptr;
		}
	return *this;
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Read a bitmap image from a file.
//!
//! \param	Path					Pathname of the file.
//!
//! \throw	FileExceptionType::FILE_OPEN_ERROR	If an error occurs while
//!							opening the file.
//! \throw	FileExceptionType::FILE_SEEK_ERROR	If an error occurs while
//!							seeking the file.
//! \throw	FileExceptionType::FILE_READ_ERROR	If an error occurs while
//!							reading the file.
//! \throw	FileExceptionType::FILE_CLOSE_ERROR	If an error occurs while
//!							closing the file.
//! \throw	IntegerExceptionType::INVALID_FILE_SIZE	If the file size in the
//!							header is not equal to
//!							the actual file size.
//! \throw	IntegerExceptionType::INVALID_BITMAP_OFFSET
//!							If the bitmap offset in
//!							the header is not equal
//!							to HEADER_SIZE +
//!							DIB_HEADER_SIZE.
//! \throw	IntegerExceptionType::INVALID_DIB_HEADER_SIZE
//!							If the DIB header size
//!							in the header is not
//!							equal to
//!							DIB_HEADER_SIZE.
//! \throw	IntegerExceptionType::INVALID_IMAGE_WIDTH
//!							If the image width is 0
//!							or greater than
//!							MAXIMUM_WIDTH.
//! \throw	IntegerExceptionType::INVALID_COLOR_PLANES
//!							If the number of color
//!							planes is not
//!							NUMBER_OF_COLOR_PLANES.
//! \throw	IntegerExceptionType::INVALID_BITS_PER_PIXEL
//!							If the number of bits
//!							per pixel is not
//!							NUMBER_OF_COLORS *
//!							CHAR_BIT.
//! \throw	IntegerExceptionType::INVALID_COMPRESSION
//!							If the value of
//!							compression is not
//!							VALID_COMPRESSION.
//! \throw	IntegerExceptionType::INVALID_IMAGE_SIZE
//!							If the image size is not
//!							valid.
//! \throw	IntegerExceptionType::INVALID_NUMBER_OF_COLORS
//!							If the number of colors
//!							is not zero.
//! \throw	IntegerExceptionType::INVALID_NUMBER_OF_IMPORTANT_COLORS
//!							If the number of
//!							important colors is not
//!							zero.
//!
//! \returns	None
//!
//! \note
//!	If Path is not a valid pointer to a zero terminated string, the behavior
//!	is undefined.
//!
//! \see	SHORT_DATA_SIZE
//! \see	LONG_DATA_SIZE
//! \see	HEADER_SIZE
//! \see	OFFSET_OF_FILE_SIZE
//! \see	HEADER_NAME_LENGTH
//! \see	OFFSET_OF_BITMAP_OFFSET
//! \see	HEADER_SIZE
//! \see	DIB_HEADER_SIZE
//! \see	OFFSET_OF_WIDTH
//! \see	OFFSET_OF_HEIGHT
//! \see	OFFSET_OF_COLOR_PLANES
//! \see	OFFSET_OF_BITS_PER_PIXEL
//! \see	CHAR_BIT
//! \see	OFFSET_OF_COMPRESSION
//! \see	OFFSET_OF_IMAGE_SIZE
//! \see	OFFSET_OF_HORIZONTAL_RESOLUTION
//! \see	OFFSET_OF_VERTICAL_RESOLUTION
//! \see	OFFSET_OF_NUMBER_OF_COLORS
//! \see	OFFSET_OF_NUMBER_OF_IMPORTANT_COLORS
//! \see	NUMBER_OF_COLORS
//! \see	MAXIMUM_WIDTH
//! \see	MAXIMUM_HEIGHT
//! \see	NUMBER_OF_COLOR_PLANES
//! \see	VALID_COMPRESSION
//!
////////////////////////////////////////////////////////////////////////////////

void	Bitmap::ReadFile(const char* Path) noexcept(false)
	{
	register	unsigned int	Color;
	register	uint_fast32_t	Value;
	register	uint_fast32_t	Row;
	register	uint_fast32_t	Column;
	register	IndexType	Index;
	register	FILE*		InputFile;
			int_fast32_t	FileSize;
			uint8_t		Buffer [DIB_HEADER_SIZE];

	if ((InputFile = fopen(Path, OpenReadBinary)) == nullptr)
		throw FileException(FileExceptionType::FILE_OPEN_ERROR, Path);
	if (fseek(InputFile, 0, SEEK_END) < 0)
		throw FileException(FileExceptionType::FILE_SEEK_ERROR, Path);
	if ((FileSize = ftell(InputFile)) < 0)
		throw FileException(FileExceptionType::FILE_SEEK_ERROR, Path);
	if (fseek(InputFile, 0, SEEK_SET) < 0)
		throw FileException(FileExceptionType::FILE_SEEK_ERROR, Path);
	if (fread(Buffer, 1, HEADER_SIZE, InputFile) != HEADER_SIZE)
		throw FileException(FileExceptionType::FILE_READ_ERROR, Path);
	if (static_cast<uint_fast32_t>(FileSize) !=
		LoadInteger(Buffer + OFFSET_OF_FILE_SIZE, LONG_DATA_SIZE))
		throw IntegerException(IntegerExceptionType::INVALID_FILE_SIZE,
			static_cast<uint_fast32_t>(FileSize));
	if (memcmp(Buffer, HeaderName, HEADER_NAME_LENGTH) != 0)
		throw NameException(Buffer);
	if ((Value = LoadInteger(Buffer + OFFSET_OF_BITMAP_OFFSET,
		LONG_DATA_SIZE)) != HEADER_SIZE + DIB_HEADER_SIZE)
		throw IntegerException(IntegerExceptionType::
			INVALID_BITMAP_OFFSET, Value);
	if (fread(Buffer, 1, DIB_HEADER_SIZE, InputFile) != DIB_HEADER_SIZE)
		throw FileException(FileExceptionType::FILE_READ_ERROR, Path);
	switch (Value = LoadInteger(Buffer, LONG_DATA_SIZE))
		{
		case DIB_HEADER_SIZE:
			break;
		default:
			throw IntegerException(IntegerExceptionType::
				INVALID_DIB_HEADER_SIZE, Value);
		}
	Width = LoadInteger(Buffer + OFFSET_OF_WIDTH, LONG_DATA_SIZE);
	if (Width == 0 or Width > MAXIMUM_WIDTH)
		throw IntegerException(IntegerExceptionType::
			INVALID_IMAGE_WIDTH, Width);
	Height = LoadInteger(Buffer + OFFSET_OF_HEIGHT, LONG_DATA_SIZE);
	if (Height == 0 or Height > MAXIMUM_WIDTH)
		throw IntegerException(IntegerExceptionType::
			INVALID_IMAGE_HEIGHT, Height);
	BitmapSize = Width * Height;
	if (BitmapSize > Bitmap::MAXIMUM_BITMAP_SIZE)
		throw IntegerException(IntegerExceptionType::BITMAP_TOO_LARGE,
			BitmapSize);
	if ((Value = LoadInteger(Buffer + OFFSET_OF_COLOR_PLANES,
		SHORT_DATA_SIZE)) != NUMBER_OF_COLOR_PLANES)
		throw IntegerException(IntegerExceptionType::
			INVALID_COLOR_PLANES, Value);
	if ((BitsPerPixel = static_cast<uint16_t>(LoadInteger(Buffer +
		OFFSET_OF_BITS_PER_PIXEL, SHORT_DATA_SIZE))) !=
		NUMBER_OF_COLORS * CHAR_BIT)
		throw IntegerException(IntegerExceptionType::
			INVALID_BITS_PER_PIXEL, BitsPerPixel);
	if ((Index = LoadInteger(Buffer + OFFSET_OF_COMPRESSION,
		LONG_DATA_SIZE)) != VALID_COMPRESSION)
		throw IntegerException(IntegerExceptionType::
			INVALID_COMPRESSION, Index);
	if ((Value = LoadInteger(Buffer + OFFSET_OF_IMAGE_SIZE,
		LONG_DATA_SIZE)) != ComputeLineSize() * Height)
		if (Index != 0 or Value != 0)
			throw IntegerException(IntegerExceptionType::
				INVALID_IMAGE_SIZE, Value);
	HorizontalResolution = LoadInteger(Buffer +
		OFFSET_OF_HORIZONTAL_RESOLUTION, LONG_DATA_SIZE);
	VerticalResolution = LoadInteger(Buffer + OFFSET_OF_VERTICAL_RESOLUTION,
		LONG_DATA_SIZE);
	if ((Value = LoadInteger(Buffer + OFFSET_OF_NUMBER_OF_COLORS,
		LONG_DATA_SIZE)) != 0)
		throw IntegerException(IntegerExceptionType::
			INVALID_NUMBER_OF_COLORS, Value);
	if ((Value = LoadInteger(Buffer + OFFSET_OF_NUMBER_OF_IMPORTANT_COLORS,
		LONG_DATA_SIZE)) != 0)
		throw IntegerException(IntegerExceptionType::
			INVALID_NUMBER_OF_IMPORTANT_COLORS, Value);
	DestroyImage();
	AllocateImage();
	Value = ComputeLineSize();
	Index = 0;
	for (Row = 0; Row < Height; ++Row)
		{
		for (Column = 0; Column < Width; ++Column)
			{
			if (fread(Buffer, 1, NUMBER_OF_COLORS, InputFile) !=
				NUMBER_OF_COLORS)
				throw FileException(FileExceptionType::
					FILE_READ_ERROR, Path);
			for (Color = 0; Color < NUMBER_OF_COLORS; ++Color)
				PixelPointer [Color] [Index] = Buffer [Color];
			++Index;
			}
		if (Column * NUMBER_OF_COLORS < Value)
			if (fread(Buffer, 1, Value - Column * NUMBER_OF_COLORS,
				InputFile) != Value - Column * NUMBER_OF_COLORS)
				throw FileException(FileExceptionType::
					FILE_READ_ERROR, Path);
		}
	if (fclose(InputFile) != 0)
		throw FileException(FileExceptionType::FILE_CLOSE_ERROR, Path);
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Write a bitmap image to a file.
//!
//! \param	Path					Pathname of the file.
//!
//! \throw	FileExceptionType::FILE_OPEN_ERROR	If an error occurs while
//!							opening the file.
//! \throw	FileExceptionType::FILE_WRITE_ERROR	If an error occurs while
//!							writing the file.
//! \throw	FileExceptionType::FILE_CLOSE_ERROR	If an error occurs while
//!							closing the file.
//!
//! \returns	None
//!
//! \note
//!	If Path is not a valid pointer to a zero terminated string, the behavior
//!	is undefined.
//!
//! \see	SHORT_DATA_SIZE
//! \see	LONG_DATA_SIZE
//! \see	HEADER_SIZE
//! \see	OFFSET_OF_FILE_SIZE
//! \see	HEADER_NAME_LENGTH
//! \see	OFFSET_OF_BITMAP_OFFSET
//! \see	HEADER_SIZE
//! \see	DIB_HEADER_SIZE
//! \see	OFFSET_OF_WIDTH
//! \see	OFFSET_OF_HEIGHT
//! \see	OFFSET_OF_COLOR_PLANES
//! \see	OFFSET_OF_BITS_PER_PIXEL
//! \see	OFFSET_OF_COMPRESSION
//! \see	OFFSET_OF_HORIZONTAL_RESOLUTION
//! \see	OFFSET_OF_VERTICAL_RESOLUTION
//! \see	NUMBER_OF_COLORS
//! \see	NUMBER_OF_COLOR_PLANES
//! \see	VALID_COMPRESSION
//!
////////////////////////////////////////////////////////////////////////////////

void	Bitmap::WriteFile(const char* Path) noexcept(false)
	{
	register	unsigned int	Color;
	register	uint_fast32_t	Length;
	register	uint_fast32_t	Row;
	register	uint_fast32_t	Column;
	register	IndexType	Index;
	register	FILE*		OutputFile;
			uint8_t		Buffer [DIB_HEADER_SIZE];

	if ((OutputFile = fopen(Path, OpenWriteBinary)) == nullptr)
		throw FileException(FileExceptionType::FILE_OPEN_ERROR, Path);
	IgnoreValue(memset(Buffer, 0, HEADER_SIZE));
	IgnoreValue(memcpy(Buffer, HeaderName, HEADER_NAME_LENGTH));
	Length = ComputeLineSize();
	StoreInteger(Buffer + OFFSET_OF_FILE_SIZE, HEADER_SIZE +
		DIB_HEADER_SIZE + Length * Height, LONG_DATA_SIZE);
	StoreInteger(Buffer + OFFSET_OF_BITMAP_OFFSET,
		HEADER_SIZE + DIB_HEADER_SIZE, LONG_DATA_SIZE);
	if (fwrite(Buffer, 1, HEADER_SIZE, OutputFile) != HEADER_SIZE)
		throw FileException(FileExceptionType::FILE_WRITE_ERROR, Path);
	IgnoreValue(memset(Buffer, 0, DIB_HEADER_SIZE));
	StoreInteger(Buffer, DIB_HEADER_SIZE, LONG_DATA_SIZE);
	StoreInteger(Buffer + OFFSET_OF_WIDTH, Width, LONG_DATA_SIZE);
	StoreInteger(Buffer + OFFSET_OF_HEIGHT, Height, LONG_DATA_SIZE);
	StoreInteger(Buffer + OFFSET_OF_COLOR_PLANES, NUMBER_OF_COLOR_PLANES,
		SHORT_DATA_SIZE);
	StoreInteger(Buffer + OFFSET_OF_BITS_PER_PIXEL, BitsPerPixel,
		SHORT_DATA_SIZE);
	StoreInteger(Buffer + OFFSET_OF_COMPRESSION, VALID_COMPRESSION,
		LONG_DATA_SIZE);
	StoreInteger(Buffer + OFFSET_OF_HORIZONTAL_RESOLUTION,
		HorizontalResolution, LONG_DATA_SIZE);
	StoreInteger(Buffer + OFFSET_OF_VERTICAL_RESOLUTION, VerticalResolution,
		LONG_DATA_SIZE);
	if (fwrite(Buffer, 1, DIB_HEADER_SIZE, OutputFile) != DIB_HEADER_SIZE)
		throw FileException(FileExceptionType::FILE_WRITE_ERROR, Path);
	Index = 0;
	assert(Length >= Width * NUMBER_OF_COLORS);
	Length -= Width * NUMBER_OF_COLORS;
	for (Row = 0; Row < Height; ++Row)
		{
		for (Column = 0; Column < Width; ++Column)
			{
			for (Color = 0; Color < NUMBER_OF_COLORS; ++Color)
				Buffer [Color] = PixelPointer [Color] [Index];
			if (fwrite(Buffer, 1, NUMBER_OF_COLORS, OutputFile) !=
				NUMBER_OF_COLORS)
				throw FileException(FileExceptionType::
					FILE_WRITE_ERROR, Path);
			++Index;
			}
		if (Length > 0)
			{
			IgnoreValue(memset(Buffer, 0, Length));
			if (fwrite(Buffer, 1, Length, OutputFile) != Length)
				throw FileException(FileExceptionType::
					FILE_WRITE_ERROR, Path);
			}
		}
	if (fclose(OutputFile) != 0)
		throw FileException(FileExceptionType::FILE_CLOSE_ERROR, Path);
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Correct the bitmap.
//!
//! \param	Parameters			Reference to the struct
//!						containing the correction
//!						parameters.
//!
//! \throw	IntegerExceptionType::INVALID_NUMBER_OF_TERMITES
//!						If the number of termites is 0
//!						or greater than
//!						MAXIMUM_NUMBER_OF_TERMITES.
//! \throw	IntegerExceptionType::INVALID_NUMBER_OF_STEPS
//!						If the number of steps is 0 or
//!						greater than
//!						MAXIMUM_NUMBER_OF_STEPS.
//! \throw	FloatExceptionType::INVALID_REPULSION_STEP
//!						If the repulsion step is less
//!						than 0 or greater than 1.
//! \throw	FloatExceptionType::INVALID_ALPHA
//!						If Alpha is less than 0 or
//!						greater than 1.
//! \throw	FloatExceptionType::INVALID_BETA
//!						If Beta is less than 0 or
//!						greater than 1.
//!
//! \returns	Bitmap				The corrected bitmap.
//!
//! \see	NUMBER_OF_COLORS
//! \see	MAXIMUM_NUMBER_OF_TERMITES
//! \see	MAXIMUM_NUMBER_OF_STEPS
//! \see	ATTRACTION_SCALE
//! \see	REPULSION_SCALE
//! \see	REPULSION_TABLE_SIZE
//! \see	UINT8_MAX
//! \see	TERMITES_SUM_SCALE
//!
////////////////////////////////////////////////////////////////////////////////

Bitmap	Bitmap::Correct(register const CorrectionParameters& Parameters)
		noexcept(false)
	{
	register	unsigned int		FirstValue;
	register	unsigned int		SecondValue;
	register	IndexType		Index;
	register	float			Sqrt2;
	register	PixelInformationType*	PixelInformationPointer;
	register	PathFunction		PathFinder;
	register	RepulsionType		InitialRepulsion;
			Bitmap			LocalBitmap(*this);
			uint_least32_t		OrthogonalAttraction
							[UINT8_MAX + 1];
			uint_least32_t		DiagonalAttraction
							[UINT8_MAX + 1];
			uint_least32_t		DivisionTable [UINT8_MAX + 1];

	if (Parameters.Termites == 0 or Parameters.Termites >
		MAXIMUM_NUMBER_OF_TERMITES)
		throw IntegerException(IntegerExceptionType::
			INVALID_NUMBER_OF_TERMITES, Parameters.Termites);
	if (Parameters.Steps == 0 or Parameters.Steps > MAXIMUM_NUMBER_OF_STEPS)
		throw IntegerException(IntegerExceptionType::
			INVALID_NUMBER_OF_STEPS, Parameters.Steps);
	if (Parameters.RepulsionStep < 0.0 or Parameters.RepulsionStep > 1.0)
		throw FloatException(FloatExceptionType::INVALID_REPULSION_STEP,
			Parameters.RepulsionStep);
	if (Parameters.Alpha < 0.0 or Parameters.Alpha > 1.0)
		throw FloatException(FloatExceptionType::INVALID_ALPHA,
			Parameters.Alpha);
	if (Parameters.Beta < 0.0 or Parameters.Beta > 1.0)
		throw FloatException(FloatExceptionType::INVALID_BETA,
			Parameters.Beta);
	PathFinder = Parameters.ChoiceKind == ChoiceType::RANDOM?
		&PixelInformationType::NextPixelRandom:
		&PixelInformationType::NextPixelMaximum;
	for (Index = 0; Index < NumberOfElements(RepulsionTable); ++Index)
		RepulsionTable [Index] = static_cast<unsigned short>
			(lrintf(powf(static_cast<float>(1) / (1 +
			static_cast<float>(Index) * Parameters.RepulsionStep),
			Parameters.Alpha) * REPULSION_SCALE));
	Sqrt2 = sqrtf(2);
	for (Index = 0; Index < NumberOfElements(OrthogonalAttraction); ++Index)
		{
		OrthogonalAttraction [Index] = static_cast<uint_least32_t>
			(lrintf(powf(static_cast<float>(1 + Index) / Sqrt2,
			Parameters.Beta) * ATTRACTION_SCALE));
		DiagonalAttraction [Index] = static_cast<uint_least32_t>
			(lrintf(powf((Sqrt2 + static_cast<float>(Index)) /
			Sqrt2, Parameters.Beta) * ATTRACTION_SCALE));
		}
	for (Index = 1; Index < NumberOfElements(DivisionTable); ++Index)
		DivisionTable [Index] = static_cast<uint_least32_t>
			(RoundedDivide(static_cast<uint_least32_t>(UINT8_MAX) *
			TERMITES_SUM_SCALE, Index));
	DivisionTable [0] = DivisionTable [1];
	PixelInformationPointer = new PixelInformationType [BitmapSize];
	InitialRepulsionPointer = new RepulsionType [BitmapSize];
	InitialRepulsion.Visitors = 0;
	InitialRepulsion.Repulsion = RepulsionTable [0];
	for (Index = 0; Index < BitmapSize; ++Index)
		InitialRepulsionPointer [Index] = InitialRepulsion;
	for (FirstValue = 0; FirstValue < Height; ++FirstValue)
		for (SecondValue = 0; SecondValue < Width; ++SecondValue)
			PixelInformationPointer [MakeIndex(FirstValue,
				SecondValue)].
				SetLinks(*this, FirstValue, SecondValue);
	for (FirstValue = 0; FirstValue < NUMBER_OF_COLORS; ++FirstValue)
		{
		SecondValue = Parameters.TruncatePath? 0: UINT8_MAX + 1;
		for (Index = 0; Index < BitmapSize; ++Index)
			{
			PixelInformationPointer [Index].
				SetAttraction(*this, PixelPointer [FirstValue],
				Index, OrthogonalAttraction,
				DiagonalAttraction);
			if (PixelPointer [FirstValue] [Index] > SecondValue)
				SecondValue = PixelPointer [FirstValue] [Index];
			}
		CorrectColor(LocalBitmap.PixelPointer [FirstValue],
			PixelPointer [FirstValue], SecondValue, Parameters,
			PixelInformationPointer, PathFinder, DivisionTable);
		}
	delete [] PixelInformationPointer;
	delete [] InitialRepulsionPointer;
	return LocalBitmap;
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Correct one color of the bitmap, on an horizontal strip of the bitmap.
//!
//! \param	Argument	Pointer to a CorrectionDescriptorType,
//!				containing all the information needed about the
//!				strip.
//! \throw	None.
//!
//! \returns	void*		Always nullptr.
//!
//! \note
//!	The correction algorithm is applied to all pixel whose row is in the
//!	range StartRow .. EndRow - 1, where StartRow and EndRow are the fields
//!	of the CorrectionDescriptorType struct.
//!
//! \note
//!	The interface if constrained, as this must be the starting function of
//!	a thread; the returned value is ignored.
//!
//! \note
//!	If Argument is not a pointer to a valid CorrectionDescriptorType struct,
//!	the behavior is undefined.
//!
////////////////////////////////////////////////////////////////////////////////

void*	Bitmap::CorrectFunction(void* Argument) noexcept(false)
	{
	const	CorrectionDescriptorType*	DescriptorPointer;
		Bitmap*				BitmapPointer;

//	DescriptorPointer =
//		static_cast<const CorrectionDescriptorType*>(Argument);
	BitmapPointer = static_cast<Bitmap*>(Argument);
	while ((DescriptorPointer = BitmapPointer->GetStripDescriptor()) !=
		nullptr)
		BitmapPointer->
			CorrectStrip(DescriptorPointer->CorrectedColorPointer,
			DescriptorPointer->OriginalColorPointer,
			DescriptorPointer->StartRow, DescriptorPointer->EndRow,
			DescriptorPointer->MaximumValue,
			*(DescriptorPointer->ParametersPointer),
			DescriptorPointer->PixelInformationPointer,
			DescriptorPointer->PathFinder,
			DescriptorPointer->DivisionTablePointer);
	return nullptr;
	}

}
