////////////////////////////////////////////////////////////////////////////////
//! \file
//!
//! \author	Mauro Fiorentini
//!
//! \date	20/09/2021
//!
//! \par	Function:
//!	Contains the definitions of non inline functions of the
//!	GraphicSpace::FileException class.
//!
////////////////////////////////////////////////////////////////////////////////

# include	"Graphic.h"
# include	"FileException.h"

namespace	GraphicSpace
{

const	char*	const	FileException::ErrorFormat = "I/O error: %s file %s\n";

const	char*	const	FileException::Message [] =
	{
	"cannot open",
	"error while reading from",
	"error while writing on",
	"error while seeking on",
	"error while closing"
	};

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Constructor.
//!
//! \param	thisExceptionKind	Kind of the exception.
//! \param	thisFileName		Name of the file that caused the
//!					exception.
//!
//! \throw	None.
//!
//! \returns	Not applicable
//!
//! \note
//!	If thisFileName is not a valid pointer to a zero terminated string, the
//!	behavior is unpredictable.
//!
//! \note
//!	The file name must be valid at the moment in which the exception is
//!	caught and a message is printed, otherwise the behavior is
//!	unpredictable.
//!
////////////////////////////////////////////////////////////////////////////////

FileException::FileException(FileExceptionType thisExceptionKind,
	const char* thisFileName) noexcept: Exception(),
	ExceptionKind(thisExceptionKind), FileName(thisFileName)
	{
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Write an error message.
//!
//! \param	OutputFile	Pointer to the file on which the message will be
//!				written.
//!
//! \throw	None.
//!
//! \returns	None.
//!
//! \note
//!	If OutputFile is not a valid file, opened for writing, the behavior is
//!	unpredictable.
//!
//! \note
//!	If an error occurs while writing on OutputFile, the process is
//!	terminated.
//!
////////////////////////////////////////////////////////////////////////////////

void	FileException::Write(FILE* OutputFile) const noexcept
	{
	if (fprintf(OutputFile, ErrorFormat,
		Message [static_cast<unsigned int>(ExceptionKind)], FileName) <
		0)
		FatalError();
	}

}
