////////////////////////////////////////////////////////////////////////////////
//! \file
//!
//! \author	Mauro Fiorentini
//!
//! \date	18/10/2021
//!
//! \par	Function:
//!	This file contains the definition of the specialization of the non
//!	inline functions of the GraphicSpace::NumericParameterDescriptor class
//!	template, with NUMERIC_TYPE equal to float and EXCEPTION_TYPE equal to
//!	FloatExceptionType.
//!
////////////////////////////////////////////////////////////////////////////////

# include	"Graphic.h"
# include	"NumericParameterDescriptor.h"
# include	"FloatException.h"

# include	<string.h>
# include	<ctype.h>
# include	<errno.h>

namespace	GraphicSpace
{

//!	Template specialization valid when NUMERIC_TYPE is float.
template<>
const	char*	NumericParameterDescriptor<float,
			FloatExceptionType>::UsageFormat =
	"\t%s%c<number> with number in the range [%g .. %g] (default %g)\n";

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Convert a string to an integer value and assign it to a parameter. \n
//!	Template specialization valid when NUMERIC_TYPE is unsigned int and
//!	EXCEPTION_TYPE is IntegerExceptionType.
//!
//! \param	String				Pointer to the string containing
//!						the value of the parameter.
//! \param	Parameters			Reference to the
//!						CorrectionParameter struct to
//!						modify with the new value.
//!
//! \throw	FloatException::ExceptionKind	If String does not begin with an
//!						equal sign or the rest does not
//!						contain a valid real number
//!						without leading sign or extra
//!						characters.
//!
//! \returns	None.
//!
//! \note
//!	if String is not a valid pointer to a properly terminated string, the
//!	behavior is undefined.
//!
//! \see	EQUAL_CHARACTER
//! \see	END_STRING
//!
////////////////////////////////////////////////////////////////////////////////

template<>
void	NumericParameterDescriptor<float, FloatExceptionType>::Assign(const
		char* String, Bitmap::CorrectionParameters& Parameters) const
		noexcept(false)
	{
	char*	EndPointer;
	float	Value;

	if (*String != EQUAL_CHARACTER)
		throw FloatException(ExceptionKind, 0);
	if (not isdigit(*++String))
		throw FloatException(ExceptionKind, 0);
	errno = 0;
	Value = strtof(String, &EndPointer);
	if (*EndPointer != END_STRING or errno != 0 or Value < Minimum or
		Value > Maximum)
		throw FloatException(ExceptionKind, Value);
	Parameters.*ParameterPointer = Value;
	}

}
