////////////////////////////////////////////////////////////////////////////////
//! \file
//!
//! \author	Mauro Fiorentini
//!
//! \date	17/10/2021
//!
//! \par	Function:
//!	This file contains the definition of the specialization of the non
//!	inline functions of the GraphicSpace::NumericParameterDescriptor class
//!	template, with NUMERIC_TYPE equal to unsigned int and EXCEPTION_TYPE
//!	equal to IntegerExceptionType.
//!
////////////////////////////////////////////////////////////////////////////////

# include	"Graphic.h"
# include	"NumericParameterDescriptor.h"
# include	"IntegerException.h"

# include	<string.h>
# include	<ctype.h>
# include	<errno.h>

namespace	GraphicSpace
{

//!	Template specialization valid when NUMERIC_TYPE is unsigned int.
template<>
const	char*	NumericParameterDescriptor<unsigned int,
			IntegerExceptionType>::UsageFormat =
	"\t%s%c<number> with number in the range [%u .. %u] (default %u)\n";

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Convert a string to an integer value and assign it to a parameter. \n
//!	Template specialization valid when NUMERIC_TYPE is unsigned int and
//!	EXCEPTION_TYPE is IntegerExceptionType.
//!
//! \param	String				Pointer to the string containing
//!						the value of the parameter.
//! \param	Parameters			Reference to the
//!						CorrectionParameter struct to
//!						modify with the new value.
//!
//! \throw	IntegerException::ExceptionKind	If String does not begin with an
//!						equal sign or the rest is not
//!						composed only of digits or if it
//!						represents a number less than
//!						Minimum or greater than Maximum.
//!
//! \returns	None.
//!
//! \note
//!	if String is not a valid pointer to a properly terminated string, the
//!	behavior is undefined.
//!
//! \see	EQUAL_CHARACTER
//! \see	DECIMAL_BASE
//! \see	END_STRING
//! \see	UINT_MAX
//!
////////////////////////////////////////////////////////////////////////////////

template<>
void	NumericParameterDescriptor<unsigned int, IntegerExceptionType>::
		Assign(const char* String,
		Bitmap::CorrectionParameters& Parameters) const noexcept(false)
	{
	char*		EndPointer;
	unsigned long	Value;

	if (*String != EQUAL_CHARACTER)
		throw IntegerException(ExceptionKind, 0);
	if (not isdigit(*++String))
		throw IntegerException(ExceptionKind, 0);
	errno = 0;
	Value = strtoul(String, &EndPointer, DECIMAL_BASE);
	if (*EndPointer != END_STRING or errno != 0 or Value < Minimum or
		Value > Maximum)
		throw IntegerException(ExceptionKind, Value);
	Parameters.*ParameterPointer = static_cast<unsigned int>(Value);
	}

}
