# include	"Graphic.h"
# include	"Bitmap.h"
# include	"RandomGenerator.h"
# include	"FileException.h"
# include	"IntegerException.h"
# include	"FloatException.h"
# include	"NameException.h"
# include	"ParameterDescriptor.h"
# include	"NumericParameterDescriptor.h"
# include	"EnumeratedParameterDescriptor.h"
# include	"Thread.h"

# include	<stdlib.h>
# include	<stdio.h>
# include	<string.h>
# include	<ctype.h>
# include	<errno.h>
# include	<time.h>
# include	<unistd.h>
# include	<math.h>

using	namespace	GraphicSpace;

//! \var	MAJOR_VERSION_NUMBER
//!	Major version number.
static	constexpr	unsigned int	MAJOR_VERSION_NUMBER = 1;

//! \var	MINOR_VERSION_NUMBER
//!	Minor version number.
static	constexpr	unsigned int	MINOR_VERSION_NUMBER = 1;

//! \var	OpenWriteText
//!	String used to specify text write only mode when opening a file.
static	constexpr	const	char*	const	OpenWriteText = "w";

//! \var	IMAGE_SIZE_FORMAT 
//!	Format string used to write the width and height of the bitmap.
static	constexpr	const	char*	IMAGE_SIZE_FORMAT =
	"Width = %" PRIuFAST32 ", Height = %" PRIuFAST32 ", Size = %" PRIuFAST32
	"\n";

//! \var	BITS_PER_PIXEL_FORMAT 
//!	Format string used to write the number of bits per pixel of the bitmap.
static	constexpr	const	char*	BITS_PER_PIXEL_FORMAT =
	"Bits per pixel = %u\n";

//! \var	FILE_SIZE_FORMAT 
//!	Format string used to write the file size of the bitmap.
static	constexpr	const	char*	FILE_SIZE_FORMAT = "File size = %lu\n";

//! \var	RESOLUTION_FORMAT 
//!	Format string used to write the horizontal and vertical resolution of
//!	the bitmap.
static	constexpr	const	char*	RESOLUTION_FORMAT =
	"Horizontal resolution = %" PRIuFAST32 ", Vertical resolution = %"
	PRIuFAST32 "\n";

//! \var	VERSION_FORMAT 
//!	Format string used to write the version number.
static	constexpr	const	char*	VERSION_FORMAT =
	"Termite-Retinex correction program, version %u.%u\n\n";

//! \var	USAGE_FORMAT 
//!	Format string used to write the usage message.
static	constexpr	const	char*	USAGE_FORMAT =
	"Usage: Graphic <Input Pathname> <Output Pathname> { <Options> }\n"
	"where Options can be:\n";

//! \var	INVALID_OPTION_FORMAT 
//!	Format string used to write the usage message.
static	constexpr	const	char*	INVALID_OPTION_FORMAT =
	"Invalid option %s\n";

//! \var	READING_FORMAT 
//!	Format string used to write the message that declares that reading the
//!	input file is in progress.
static	constexpr	const	char*	READING_FORMAT = "Reading %s\n";

//! \var	END_READING_FORMAT 
//!	Format string used to write the message that declares that reading the
//!	input file is terminated.
static	constexpr	const	char*	END_READING_FORMAT =
	"End of reading; elapsed time = %lu\n";

//! \var	CORRECTING_MESSAGE 
//!	Message that declares that image correction started.
static	constexpr	const	char*	CORRECTING_MESSAGE =
	"Correcting the image with %u thread(s)\n";

//! \var	END_CORRECTION_FORMAT
//!	Format string used to write the message that declares that the
//!	correction is terminated.
static	constexpr	const	char*	END_CORRECTION_FORMAT =
	"End of correction; elapsed time = %lu\n";

//! \var	WRITING_FORMAT 
//!	Format string used to write the message that declares that writing the
//!	output file is in progress.
static	constexpr	const	char*	WRITING_FORMAT = "Writing %s\n";

//! \var	END_WRITING_FORMAT 
//!	Format string used to write the message that declares that writing the
//!	output file is terminated.
static	constexpr	const	char*	END_WRITING_FORMAT =
	"End of writing; elapsed time = %lu\n";

//! \var	DefaultParameters
//!	Correction parameters, initialized with default values.
static	const	Bitmap::CorrectionParameters	DefaultParameters =
	{
	1, 100, 100, 1.0, 0.5, 0.5, false, Bitmap::ChoiceType::MAXIMUM, false,
	false
	};

//! \var	DescriptorTable
//!	Table of descriptors of correction parameters.
//!
//! \note
//!	The order of the parameters correspond to the order in which they are
//!	printed when describing the usage of the program.
const	ParameterDescriptor*	const	DescriptorTable [] =
	{
	new NumericParameterDescriptor<unsigned int, IntegerExceptionType>
		("-Threads", 1, Bitmap::MAXIMUM_NUMBER_OF_THREADS,
		&Bitmap::CorrectionParameters::Threads,
		IntegerExceptionType::INVALID_NUMBER_OF_THREADS),
	new NumericParameterDescriptor<unsigned int, IntegerExceptionType>
		("-Termites", 1, Bitmap::MAXIMUM_NUMBER_OF_TERMITES,
		&Bitmap::CorrectionParameters::Termites,
		IntegerExceptionType::INVALID_NUMBER_OF_TERMITES),
	new NumericParameterDescriptor<unsigned int, IntegerExceptionType>
		("-Steps", 1, Bitmap::MAXIMUM_NUMBER_OF_STEPS,
		&Bitmap::CorrectionParameters::Steps,
		IntegerExceptionType::INVALID_NUMBER_OF_STEPS),
	new NumericParameterDescriptor<float, FloatExceptionType>("-Q", 0.0,
		1.0, &Bitmap::CorrectionParameters::RepulsionStep,
		FloatExceptionType::INVALID_REPULSION_STEP),
	new NumericParameterDescriptor<float, FloatExceptionType>("-Alpha", 0.0,
		1.0, &Bitmap::CorrectionParameters::Alpha,
		FloatExceptionType::INVALID_REPULSION_STEP),
	new NumericParameterDescriptor<float, FloatExceptionType>("-Beta", 0.0,
		1.0, &Bitmap::CorrectionParameters::Beta,
		FloatExceptionType::INVALID_REPULSION_STEP),
	new EnumeratedParameterDescriptor<bool>("-Randomize",
		&Bitmap::CorrectionParameters::RandomizeFlag,
		EnumeratedExceptionType::INVALID_RANDOMIZE_PARAMETER),
	new EnumeratedParameterDescriptor<Bitmap::ChoiceType>("-PathChoice",
		&Bitmap::CorrectionParameters::ChoiceKind,
		EnumeratedExceptionType::INVALID_PATH_CHOICE_PARAMETER),
	new EnumeratedParameterDescriptor<bool>("-TruncatePath",
		&Bitmap::CorrectionParameters::TruncatePath,
		EnumeratedExceptionType::INVALID_TRUNCATE_PATH_PARAMETER),
	new EnumeratedParameterDescriptor<bool>("-Quiet",
		&Bitmap::CorrectionParameters::Quiet,
		EnumeratedExceptionType::INVALID_QUIET_PARAMETER)
	};

static	void	Randomize(const Bitmap::CorrectionParameters& Parameters,
			Bitmap& CurrentBitmap) noexcept;

static	void	WriteProperties(const Bitmap& CurrentBitmap, const char* Path)
			noexcept(false);

static	void	Usage(void) noexcept(false);

static	void	Application(const char* InputPathname,
			const char* OutputPathname,
			const Bitmap::CorrectionParameters& Parameters)
			noexcept(false);

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Create a set of values to initialize a random numbers generator with a
//!	non reproducible value.
//!
//! \param	Parameters	Reference to the struct containing the
//!				correction parameters.
//! \param	CurrentBitmap	Reference to the Bitmap object that will correct
//!				the image.
//!
//! \throw	None.
//!
//! \returns	None
//!
////////////////////////////////////////////////////////////////////////////////

static	void	Randomize(const Bitmap::CorrectionParameters& Parameters,
			Bitmap& CurrentBitmap) noexcept
	{
	register	unsigned int	Index;
			time_t		Time;
			tm		TimeData;

	Time = time(nullptr);
	TimeData = *gmtime(&Time);

	uint_fast32_t	Seed [RandomGenerator::NUMBER_OF_GENERATORS + 1] =
		{
		0,
		static_cast<uint_fast32_t>(TimeData.tm_sec),
		static_cast<uint_fast32_t>(TimeData.tm_min),
		static_cast<uint_fast32_t>(TimeData.tm_hour),
		static_cast<uint_fast32_t>(TimeData.tm_mday),
		static_cast<uint_fast32_t>(TimeData.tm_mon),
		static_cast<uint_fast32_t>(TimeData.tm_year),
		static_cast<uint_fast32_t>(TimeData.tm_wday),
		static_cast<uint_fast32_t>(TimeData.tm_yday),
		static_cast<uint_fast32_t>(clock()),
		static_cast<uint_fast32_t>(time(nullptr)),
		static_cast<uint_fast32_t>(getpid()),
		static_cast<uint_fast32_t>(getppid()),
		static_cast<uint_fast32_t>(getuid()),
		static_cast<uint_fast32_t>(getgid()),
		static_cast<uint_fast32_t>(reinterpret_cast<intptr_t>(Seed)),
		static_cast<uint_fast32_t>(reinterpret_cast<intptr_t>(stdin)),
		static_cast<uint_fast32_t>(reinterpret_cast<intptr_t>
			(Randomize)),
		static_cast<uint_fast32_t>(reinterpret_cast<intptr_t>(&Time)),
		static_cast<uint_fast32_t>(reinterpret_cast<intptr_t>
			(&DefaultParameters)),
		static_cast<uint_fast32_t>(Parameters.Termites),
		static_cast<uint_fast32_t>(Parameters.Steps),
		static_cast<uint_fast32_t>(lrintf(Parameters.RepulsionStep *
			static_cast<float>(UINT_FAST32_MAX))),
		static_cast<uint_fast32_t>(lrintf(Parameters.Alpha *
			static_cast<float>(UINT_FAST32_MAX))),
		static_cast<uint_fast32_t>(lrintf(Parameters.Beta *
			static_cast<float>(UINT_FAST32_MAX))),
		static_cast<uint_fast32_t>(CurrentBitmap.GetWidth()),
		static_cast<uint_fast32_t>(CurrentBitmap.GetHeight()),
		1,
		2,
		3
		};
	for (Index = 1; Index < NumberOfElements(Seed); ++Index)
		Seed [0] ^= Seed [Index];
	for (Index = 1; Index < NumberOfElements(Seed); ++Index)
		Seed [Index] ^= Seed [0] ^ Index;
	CurrentBitmap.Randomize(Seed);
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Write the properties of a bitmap image.
//!
//! \param	CurrentBitmap				Reference to the bitmap.
//! \param	Path					Pathname of the file.
//!
//! \throw	FileExceptionType::FILE_OPEN_ERROR	If an error occurs while
//!							opening the file.
//! \throw	FileExceptionType::FILE_WRITE_ERROR	If an error occurs while
//!							writing the file.
//! \throw	FileExceptionType::FILE_CLOSE_ERROR	If an error occurs while
//!							closing the file.
//!
//! \returns	None
//!
//! \note
//!	If Path is not a valid pointer to a zero terminated string, the behavior
//!	is undefined.
//!
//! \see	IMAGE_SIZE_FORMAT 
//! \see	BITS_PER_PIXEL_FORMAT 
//! \see	FILE_SIZE_FORMAT 
//! \see	RESOLUTION_FORMAT 
//!
////////////////////////////////////////////////////////////////////////////////

static	void	WriteProperties(const Bitmap& CurrentBitmap, const char* Path)
			noexcept(false)
	{
	FILE*	OutputFile;

	if (Path == nullptr)
		OutputFile = stdout;
	else
		if ((OutputFile = fopen(Path, OpenWriteText)) == nullptr)
			throw FileException(FileExceptionType::FILE_OPEN_ERROR,
				Path);
	if (fprintf(OutputFile, IMAGE_SIZE_FORMAT, CurrentBitmap.GetWidth(),
		CurrentBitmap.GetHeight(), CurrentBitmap.GetSize()) < 0)
		throw FileException(FileExceptionType::FILE_WRITE_ERROR, Path);
	if (fprintf(OutputFile, BITS_PER_PIXEL_FORMAT,
		CurrentBitmap.GetBitsPerPixel()) < 0)
		throw FileException(FileExceptionType::FILE_WRITE_ERROR, Path);
	if (fprintf(OutputFile, FILE_SIZE_FORMAT,
		static_cast<unsigned long>(Bitmap::HEADER_SIZE +
		Bitmap::DIB_HEADER_SIZE + CurrentBitmap.ComputeLineSize() *
		CurrentBitmap.GetHeight())) < 0)
		throw FileException(FileExceptionType::FILE_WRITE_ERROR, Path);
	if (fprintf(OutputFile, RESOLUTION_FORMAT,
		CurrentBitmap.GetHorizontalResolution(),
		CurrentBitmap.GetVerticalResolution()) < 0)
		throw FileException(FileExceptionType::FILE_WRITE_ERROR, Path);
	if (Path != nullptr)
		if (fclose(OutputFile) != 0)
			throw FileException(FileExceptionType::FILE_CLOSE_ERROR,
				Path);
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Write on standard output a string describing how to use the program and
//!	stop the program.
//!
//! \throw	FileExceptionType::FILE_WRITE_ERROR
//!						If an error occurs while writing
//!						the string.
//!
//! \returns	The function does nor return.
//!
//! \see	STANDARD_OUTPUT
//! \see	USAGE_FORMAT 
//!
////////////////////////////////////////////////////////////////////////////////

static	void	Usage(void) noexcept(false)
	{
	register	unsigned int	Index;

	if (printf(USAGE_FORMAT) < 0)
		throw FileException(FileExceptionType::FILE_WRITE_ERROR,
			STANDARD_OUTPUT);
	for (Index = 0; Index < NumberOfElements(DescriptorTable); ++Index)
		DescriptorTable [Index]->PrintUsage(DefaultParameters);
	exit(EXIT_FAILURE);
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Application.
//!
//! \param	InputPathName			Pointer to the string containing
//!						the pathname of the input bitmap
//!						file.
//! \param	OutputPathName			Pointer to the string containing
//!						the pathname of the output
//!						bitmap file.
//! \param	Parameters			Reference to the struct
//!						containing the correction
//!						parameters.
//!
//! \throw	FileExceptionType::FILE_OPEN_ERROR
//!						If an error occurs while opening
//!						the input or output bitmap file.
//! \throw	FileExceptionType::FILE_SEEK_ERROR
//!						If an error occurs while seeking
//!						the input bitmap file.
//! \throw	FileExceptionType::FILE_READ_ERROR
//!						If an error occurs while reading
//!						the input bitmap file.
//! \throw	FileExceptionType::FILE_WRITE_ERROR
//!						If an error occurs while writing
//!						the output bitmap file.
//! \throw	FileExceptionType::FILE_CLOSE_ERROR
//!						If an error occurs while closing
//!						the input or output bitmap file.
//! \throw	IntegerExceptionType::INVALID_FILE_SIZE
//!						If the bitmap file size in the
//!						header of the bitmap file is not
//!						equal to the actual file size.
//! \throw	IntegerExceptionType::INVALID_BITMAP_OFFSET
//!						If the bitmap offset in the
//!						header of the bitmap file is not
//!						equal to
//!						HEADER_SIZE + DIB_HEADER_SIZE.
//! \throw	IntegerExceptionType::INVALID_DIB_HEADER_SIZE
//!						If the DIB header size in the
//!						header of the bitmap file is not
//!						equal to DIB_HEADER_SIZE.
//! \throw	IntegerExceptionType::INVALID_IMAGE_WIDTH
//!						If the image width in the bitmap
//!						file is 0 or greater than
//!						MAXIMUM_WIDTH.
//! \throw	IntegerExceptionType::INVALID_COLOR_PLANES
//!						If the number of color planes in
//!						the bitmap file is not
//!						NUMBER_OF_COLOR_PLANES.
//! \throw	IntegerExceptionType::INVALID_BITS_PER_PIXEL
//!						If the number of bits per pixel
//!						in the bitmap file is not
//!						NUMBER_OF_COLORS * CHAR_BIT.
//! \throw	IntegerExceptionType::INVALID_COMPRESSION
//!						If the value of compression in
//!						the bitmap file is not
//!						VALID_COMPRESSION.
//! \throw	IntegerExceptionType::INVALID_IMAGE_SIZE
//!						If the image size in the bitmap
//!						file is not valid.
//! \throw	IntegerExceptionType::INVALID_NUMBER_OF_COLORS
//!						If the number of color in the
//!						bitmap file is not zero.
//! \throw	IntegerExceptionType::INVALID_NUMBER_OF_IMPORTANT_COLORS
//!						If the number of important
//!						colors in the bitmap file is not
//!						zero.
//! \throw	IntegerExceptionType::INVALID_NUMBER_OF_TERMITES
//!						If the number of termites is not
//!						a valid number, it is 0 or
//!						greater than
//!						MAXIMUM_NUMBER_OF_TERMITES.
//! \throw	IntegerExceptionType::INVALID_NUMBER_OF_STEPS
//!						If the number of steps is not a
//!						valid number, it is 0 or greater
//!						than MAXIMUM_NUMBER_OF_STEPS.
//! \throw	FloatExceptionType::INVALID_REPULSION_STEP
//!						If the repulsion step is not a
//!						valid number, it is less than 0
//!						or greater than 1.
//! \throw	FloatExceptionType::INVALID_ALPHA
//!						If Alpha is is not a valid
//!						number, it less than 0 or
//!						greater than 1.
//! \throw	FloatExceptionType::INVALID_BETA
//!						If Beta is is not a valid
//!						number, it less than 0 or
//!						greater than 1.
//!
//! \returns	None.
//!
//! \note
//!	if PathName is not a valid pointer to a properly terminated string, the
//!	behavior is undefined.
//!
//! \see	STANDARD_OUTPUT
//! \see	READING_FORMAT 
//! \see	END_READING_FORMAT 
//! \see	CORRECTING_MESSAGE 
//! \see	END_CORRECTION_FORMAT
//! \see	WRITING_FORMAT 
//! \see	END_WRITING_FORMAT 
//!
////////////////////////////////////////////////////////////////////////////////

void	Application(const char* InputPathName, const char* OutputPathName,
		const Bitmap::CorrectionParameters& Parameters) noexcept(false)
	{
	time_t	StartTime;
	Bitmap	SourceBitmap;
	Bitmap	CorrectedBitmap;

	StartTime = time(nullptr);
	if (printf(READING_FORMAT, InputPathName) < 0)
		throw (FileException(FileExceptionType::FILE_WRITE_ERROR,
			STANDARD_OUTPUT));
	SourceBitmap.ReadFile(InputPathName);
	if (printf(END_READING_FORMAT,
		static_cast<unsigned long>(time(nullptr) - StartTime)) < 0)
		throw (FileException(FileExceptionType::FILE_WRITE_ERROR,
			STANDARD_OUTPUT));
	if (not Parameters.Quiet)
		WriteProperties(SourceBitmap, nullptr);
	if (printf(CORRECTING_MESSAGE, Parameters.Threads) < 0)
		throw (FileException(FileExceptionType::FILE_WRITE_ERROR,
			STANDARD_OUTPUT));
	if (Parameters.RandomizeFlag)
		Randomize(Parameters, SourceBitmap);
	CorrectedBitmap = SourceBitmap.Correct(Parameters);
	if (printf(END_CORRECTION_FORMAT,
		static_cast<unsigned long>(time(nullptr) - StartTime)) < 0)
		throw (FileException(FileExceptionType::FILE_WRITE_ERROR,
			STANDARD_OUTPUT));
	if (printf(WRITING_FORMAT, OutputPathName) < 0)
		throw (FileException(FileExceptionType::FILE_WRITE_ERROR,
			STANDARD_OUTPUT));
	CorrectedBitmap.WriteFile(OutputPathName);
	if (printf(END_WRITING_FORMAT,
		static_cast<unsigned long>(time(nullptr) - StartTime)) < 0)
		throw (FileException(FileExceptionType::FILE_WRITE_ERROR,
			STANDARD_OUTPUT));
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Main program.
//!
//! \param	argc	Number of command line arguments.
//! \param	argv	Array of command line arguments.
//!
//! \throw	None.
//!
//! \returns	int	EXIT_SUCCESS, if the program terminates normally,
//!			EXIT_FAILURE otherwise.
//!
//! \see	STANDARD_OUTPUT
//! \see	VERSION_FORMAT 
//! \see	MAJOR_VERSION_NUMBER
//! \see	MINOR_VERSION_NUMBER
//! \see	INVALID_OPTION_FORMAT 
//!
////////////////////////////////////////////////////////////////////////////////

int	main(int argc, char* argv [])
	{
	register	unsigned int		Index;
	register	size_t			Length;
	register	const	char*		Pointer;
			const	char*		InputPathName;
			const	char*		OutputPathName;
			Bitmap::CorrectionParameters
						Parameters(DefaultParameters);

	try
		{
		Thread::Initialize();
		if (printf(VERSION_FORMAT, MAJOR_VERSION_NUMBER,
			MINOR_VERSION_NUMBER) < 0)
			throw (FileException(FileExceptionType::
				FILE_WRITE_ERROR, STANDARD_OUTPUT));
		if (argc < 3)
			Usage();
		InputPathName = *++argv;
		OutputPathName = *++argv;
		while (*++argv != nullptr)
			{
			if ((Pointer = strchr(*argv,
				ParameterDescriptor::EQUAL_CHARACTER)) ==
				nullptr)
				{
				IgnoreValue(fprintf(stderr,
					INVALID_OPTION_FORMAT, *argv));
				Usage();
				break;
				}
			Length = static_cast<size_t>(Pointer - *argv);
			for (Index = 0;
				Index < NumberOfElements(DescriptorTable);
				++Index)
				{
				if (Length != strlen(DescriptorTable [Index]->
					GetName()))
					continue;
				if (strncmp(*argv, DescriptorTable [Index]->
					GetName(), Length) != 0)
					continue;
				DescriptorTable [Index]->
					Assign(Pointer, Parameters);
				break;
				}
			if (Index == NumberOfElements(DescriptorTable))
				{
				IgnoreValue(fprintf(stderr,
					INVALID_OPTION_FORMAT, *argv));
				Usage();
				break;
				}
			}
		Application(InputPathName, OutputPathName, Parameters);
		Thread::Finalize();
		}
	catch (FileException Exception)
		{
		Exception.Write(stderr);
		exit(EXIT_FAILURE);
		}
	catch (IntegerException Exception)
		{
		Exception.Write(stderr);
		Usage();
		exit(EXIT_FAILURE);
		}
	catch (FloatException Exception)
		{
		Exception.Write(stderr);
		Usage();
		exit(EXIT_FAILURE);
		}
	catch (NameException Exception)
		{
		Exception.Write(stderr);
		exit(EXIT_FAILURE);
		}
	return EXIT_SUCCESS;
	}
