////////////////////////////////////////////////////////////////////////////////
//! \file
//!
//! \author	Mauro Fiorentini
//!
//! \date	16/08/2022
//!
//! \par	Function:
//!	Contains the definitions of non inline functions of the
//!	GraphicSpace::Thread class.
//!
////////////////////////////////////////////////////////////////////////////////

# include	"Graphic.h"
# include	"SystemException.h"
# include	"Mutex.h"

namespace	GraphicSpace
{

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Constructor.
//!
//! \throw	None.
//!
//! \returns	Not applicable.
//!
////////////////////////////////////////////////////////////////////////////////

Mutex::Mutex(void) noexcept: Active(false), MutexData()
	{
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Destructor.
//!
//! \throw	SystemExceptionType::MUTEX_DESTROY_FAILED
//!					If the destruction fails.
//!
//! \returns	Not applicable.
//!
////////////////////////////////////////////////////////////////////////////////

Mutex::~Mutex(void) noexcept(false)
	{
	if (Active)
		Finalize();
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Initialize the mutex.
//!
//! \throw	SystemExceptionType::INVALID_MUTEX_OPERATION
//!					If the Mutex is already initialized.
//! \throw	SystemExceptionType::MUTEX_INITIALIZATION_FAILED
//!					If the initialization fails.
//!
//! \returns	Not applicable.
//!
//! \see	MUTEX_INIT_OK
//!
////////////////////////////////////////////////////////////////////////////////

void	Mutex::Initialize(void) noexcept(false)
	{
	if (Active)
		throw SystemException(SystemExceptionType::
			INVALID_MUTEX_OPERATION);
	if (pthread_mutex_init(&MutexData, nullptr) != MUTEX_INIT_OK)
		throw SystemException(SystemExceptionType::
			MUTEX_INITIALIZATION_FAILED);
	Active = true;
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Destroy the Mutex.
//!
//! \throw	SystemExceptionType::INVALID_MUTEX_OPERATION
//!					If the Mutex is not initialized.
//! \throw	SystemExceptionType::MUTEX_DESTROY_FAILED
//!					If the operation fails.
//!
//! \returns	Not applicable.
//!
//! \see	MUTEX_DESTROY_OK
//!
////////////////////////////////////////////////////////////////////////////////

void	Mutex::Finalize(void) noexcept(false)
	{
	if (not Active)
		throw SystemException(SystemExceptionType::
			INVALID_MUTEX_OPERATION);
	if (pthread_mutex_destroy(&MutexData) != MUTEX_DESTROY_OK)
		throw SystemException(SystemExceptionType::
			MUTEX_DESTROY_FAILED);
	Active = false;
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Lock the mutex.
//!
//! \throw	SystemExceptionType::INVALID_MUTEX_OPERATION
//!					If the Mutex is not initialized.
//! \throw	SystemExceptionType::MUTEX_LOCK_FAILED	If the operation fails.
//!
//! \returns	None.
//!
//! \see	MUTEX_LOCK_OK
//!
////////////////////////////////////////////////////////////////////////////////

void	Mutex::Lock(void) noexcept(false)
	{
	if (not Active)
		throw SystemException(SystemExceptionType::
			INVALID_MUTEX_OPERATION);
	if (pthread_mutex_lock(&MutexData) != MUTEX_LOCK_OK)
		throw SystemException(SystemExceptionType::MUTEX_LOCK_FAILED);
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Unlock the mutex.
//!
//! \throw	SystemExceptionType::INVALID_MUTEX_OPERATION
//!					If the Mutex is not initialized.
//! \throw	SystemExceptionType::MUTEX_UNLOCK_FAILED	If the operation
//!								fails.
//!
//! \returns	None.
//!
//! \see	MUTEX_UNLOCK_OK
//!
////////////////////////////////////////////////////////////////////////////////

void	Mutex::Unlock(void) noexcept(false)
	{
	if (not Active)
		throw SystemException(SystemExceptionType::
			INVALID_MUTEX_OPERATION);
	if (pthread_mutex_unlock(&MutexData) != MUTEX_UNLOCK_OK)
		throw SystemException(SystemExceptionType::MUTEX_UNLOCK_FAILED);
	}

}
