////////////////////////////////////////////////////////////////////////////////
//! \file
//!
//! \author	Mauro Fiorentini
//!
//! \date	16/08/2022
//!
//! \par	Function:
//!	Contains the definitions of non inline functions of the
//!	GraphicSpace::Thread class.
//!
////////////////////////////////////////////////////////////////////////////////

# include	"Graphic.h"
# include	"SystemException.h"
# include	"Thread.h"

# include	<signal.h>

namespace	GraphicSpace
{

pthread_attr_t	Thread::ThreadAttribute;

pthread_t	Thread::MainThreadId;

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Constructor.
//!
//! \throw	None.
//!
//! \returns	Not applicable.
//!
////////////////////////////////////////////////////////////////////////////////

Thread::Thread(void) noexcept: ThreadId(MainThreadId)
	{
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Destructor.
//!
//! \throw	None.
//!
//! \returns	Not applicable.
//!
//! \note
//!	If the Thread is running, it is killed.
//!
////////////////////////////////////////////////////////////////////////////////

Thread::~Thread(void) noexcept
	{
	if (not pthread_equal(ThreadId, MainThreadId))
		Kill();
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Initialize the thread management.
//!
//! \throw	SystemExceptionType::THREADINITIALIZATION_FAILED
//!					If the initialization fails.
//!
//! \returns	None.
//!
//! \note
//!	If this function is not called before any thread operation, including
//!	the definition of a Thread object, the behavior is undefined.
//!
//! \see	ATTRIBUTE_INIT_OK
//! \see	ATTRIBUTE_SET_OK
//! \see	THREAD_CREATE_OK
//!
////////////////////////////////////////////////////////////////////////////////

void	Thread::Initialize(void) noexcept(false)
	{
	sched_param	SchedulingParameter;

	if (pthread_attr_init(&ThreadAttribute) != ATTRIBUTE_INIT_OK)
		throw SystemException(SystemExceptionType::
			THREAD_INITIALIZATION_FAILED);
	if (pthread_attr_setschedpolicy(&ThreadAttribute, SCHED_OTHER) !=
		ATTRIBUTE_SET_OK)
		throw SystemException(SystemExceptionType::
			THREAD_INITIALIZATION_FAILED);
	if (pthread_attr_setdetachstate(&ThreadAttribute,
		PTHREAD_CREATE_JOINABLE) != ATTRIBUTE_SET_OK)
		throw SystemException(SystemExceptionType::
			THREAD_INITIALIZATION_FAILED);
	SchedulingParameter.sched_priority = 0;
	if (pthread_attr_setschedparam(&ThreadAttribute,
		&SchedulingParameter) != ATTRIBUTE_SET_OK)
		throw SystemException(SystemExceptionType::
			THREAD_INITIALIZATION_FAILED);
	if (pthread_attr_setinheritsched(&ThreadAttribute,
		PTHREAD_EXPLICIT_SCHED) != ATTRIBUTE_SET_OK)
		throw SystemException(SystemExceptionType::
			THREAD_INITIALIZATION_FAILED);
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Finalize the thread management.
//!
//! \throw	None.
//!
//! \returns	None.
//!
////////////////////////////////////////////////////////////////////////////////

void	Thread::Finalize(void) noexcept
	{
	IgnoreValue(pthread_attr_destroy(&ThreadAttribute));
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Start the thread.
//!
//! \param	ThreadFunction					Pointer to the
//!								Thread object.
//! \param	Argument					Pointer to the
//!								Thread argument.
//!
//! \throw	SystemExceptionType::THREAD_CREATION_FAILED	If the thread
//!								creation fails
//!								or if scheduling
//!								parameters
//!								setting fails.
//! \throw	SystemExceptionType::INVALID_THREAD_OPERATION	If the Thread
//!								was already
//!								started.
//!
//! \returns	None.
//!
//! \note
//!	After started, the Thread cannot be started again, unless it has
//!	terminated, or it has been killed.
//!
//! \see	THREAD_CREATE_OK
//!
////////////////////////////////////////////////////////////////////////////////

void	Thread::Start(ThreadFunctionType ThreadFunction, void* Argument)
		noexcept(false)
	{
	if (not pthread_equal(ThreadId, MainThreadId))
		throw SystemException(SystemExceptionType::
			INVALID_THREAD_OPERATION);
	if (pthread_create(&ThreadId, &ThreadAttribute, ThreadFunction,
		Argument) != THREAD_CREATE_OK)
		throw SystemException(SystemExceptionType::
			THREAD_CREATION_FAILED);
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Kill the thread.
//!
//! \throw	SystemExceptionType::INVALID_THREAD_OPERATION	If the Thread
//!								was not started.
//!
//! \returns	None.
//!
//! \note
//!	After having been killed, the Thread can be started again.
//!
////////////////////////////////////////////////////////////////////////////////

void	Thread::Kill(void) noexcept(false)
	{
	if (pthread_equal(ThreadId, MainThreadId))
		throw SystemException(SystemExceptionType::
			INVALID_THREAD_OPERATION);
	IgnoreValue(pthread_kill(ThreadId, SIGKILL));
	ThreadId = MainThreadId;
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Make the calling thread wait, until the Thread on which the operation is
//!	invoked is terminated.
//!
//! \throw	SystemExceptionType::INVALID_THREAD_OPERATION	If the Thread
//!								was not started.
//!
//! \returns	None.
//!
//! \see	THREAD_JOIN_OK
//!
////////////////////////////////////////////////////////////////////////////////

void	Thread::Join(void) noexcept(false)
	{
	if (pthread_equal(ThreadId, MainThreadId))
		throw SystemException(SystemExceptionType::
			INVALID_THREAD_OPERATION);
	if (pthread_join(ThreadId,
		static_cast<void **>(NULL)) != THREAD_JOIN_OK)
		throw SystemException(SystemExceptionType::
			INVALID_THREAD_OPERATION);
	ThreadId = MainThreadId;
	}

}
