////////////////////////////////////////////////////////////////////////////////
//! \file
//!
//! \author	Mauro Fiorentini
//!
//! \date	12/09/2021
//!
//! \par	Function:
//!	Contains the definitions of inline functions of the
//!	GraphicSpace::Bitmap class.
//!
////////////////////////////////////////////////////////////////////////////////

namespace	GraphicSpace
{

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Set the index of a neighbour of a pixel.
//!
//! \param	Image	Reference to the Bitmap containing the pixel.
//! \param	Row	Row index.
//! \param	Column	Column index.
//!
//! \throw	None.
//!
//! \returns	bool	true, if the neighbour exists.
//!
//! \note
//!	The neighbour exits iff Row < Height and Column < Width.
//!
////////////////////////////////////////////////////////////////////////////////

inline	bool	Bitmap::PixelInformationType::NeighbourType::SetLink(const
			Bitmap& Image, unsigned int Row, unsigned int Column)
			noexcept
	{
	if (not Image.ValidIndex(Row, Column))
		return false;
	Index = static_cast<uint_least32_t>(Image.MakeIndex(Row, Column));
	return true;
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Set the links to the neighbourhood of a pixel.
//!
//! \param	Image	Reference to the Bitmap containing the pixel.
//! \param	Row	Row index.
//! \param	Column	Column index.
//!
//! \throw	None.
//!
//! \returns	none.
//!
//! \note
//!	When computing the row and colum indices of a neighbour, the function
//!	exploits the modular properties of unsigned arithmetic, so subtracting
//!	1 from a 0 index generates a large value, which is greater than the
//!	number of rows and columns.
//!
////////////////////////////////////////////////////////////////////////////////

inline	void	Bitmap::PixelInformationType::SetLinks(const Bitmap& Image,
			unsigned int Row, unsigned int Column) noexcept
	{
	NumberOfNeighbours = 0;
	if (Neighbour [NumberOfNeighbours].SetLink(Image, Row - 1, Column - 1))
		++NumberOfNeighbours;
	if (Neighbour [NumberOfNeighbours].SetLink(Image, Row - 1, Column))
		++NumberOfNeighbours;
	if (Neighbour [NumberOfNeighbours].SetLink(Image, Row - 1, Column + 1))
		++NumberOfNeighbours;
	if (Neighbour [NumberOfNeighbours].SetLink(Image, Row, Column - 1))
		++NumberOfNeighbours;
	if (Neighbour [NumberOfNeighbours].SetLink(Image, Row, Column + 1))
		++NumberOfNeighbours;
	if (Neighbour [NumberOfNeighbours].SetLink(Image, Row + 1, Column - 1))
		++NumberOfNeighbours;
	if (Neighbour [NumberOfNeighbours].SetLink(Image, Row + 1, Column))
		++NumberOfNeighbours;
	if (Neighbour [NumberOfNeighbours].SetLink(Image, Row + 1, Column + 1))
		++NumberOfNeighbours;
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Set the attraction to the neighbouhhood of a pixel.
//!
//! \param	Image			Reference to the Bitmap containing the
//!					pixel.
//! \param	ColorPointer		Pointer to the array that contains the
//!					the bitmap of the color.
//! \param	PixelIndex		Index of the pixel.
//! \param	OrthogonalAttraction	Array of attraction factor for
//!					orthogonal neighbours.
//! \param	DiagonalAttraction	Array of attraction factor for diagonal
//!					neighbours.
//!
//! \throw	None.
//!
//! \returns	none.
//!
//! \note
//!	The index in OrthogonalAttraction and DiagonalAttraction is the absolute
//!	difference of intensity and the corresponding value is the attraction
//!	factor, scaled by ATTRACTION_SCALE.
//!
//! \see	UINT8_MAX
//!
////////////////////////////////////////////////////////////////////////////////

inline	void	Bitmap::PixelInformationType::SetAttraction(register const
			Bitmap& Image, register const uint8_t* ColorPointer,
			register IndexType PixelIndex,
			register const uint_least32_t
			OrthogonalAttraction [UINT8_MAX + 1], register const
			uint_least32_t DiagonalAttraction [UINT8_MAX + 1])
			noexcept
	{
	register	unsigned int	Index;

	for (Index = 0; Index < NumberOfNeighbours; ++Index)
		Neighbour [Index].Attraction =
			(Image.IsOrthogonal(Index, Neighbour [Index].Index)?
			OrthogonalAttraction: DiagonalAttraction)
		       	[abs(ColorPointer [PixelIndex] -
			ColorPointer [Neighbour [Index].Index])];
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Allocate the bitmap for the image.
//!
//! \throw	None.
//!
//! \returns	none.
//!
//! \see	NUMBER_OF_COLORS
//!
////////////////////////////////////////////////////////////////////////////////

inline	void	Bitmap::AllocateImage(void) noexcept
	{
	register	unsigned int	Color;

	for (Color = 0; Color < NUMBER_OF_COLORS; ++Color)
		PixelPointer [Color] =
			BitmapSize == 0? nullptr: new uint8_t [BitmapSize];
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Free the bitmap of the image.
//!
//! \throw	None.
//!
//! \returns	none.
//!
//! \see	NUMBER_OF_COLORS
//!
////////////////////////////////////////////////////////////////////////////////

inline	void	Bitmap::DestroyImage(void) noexcept
	{
	register	unsigned int	Color;

	for (Color = 0; Color < NUMBER_OF_COLORS; ++Color)
		if (PixelPointer [Color] != nullptr)
			delete [] PixelPointer [Color];
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Check whether the Bitmap contains the pixel.
//!
//! \param	Row	Row index.
//! \param	Column	Column index.
//!
//! \throw	None.
//!
//! \returns	bool	true, if the Bitmap contains the pixel.
//!
////////////////////////////////////////////////////////////////////////////////

inline	bool	Bitmap::ValidIndex(unsigned int Row, unsigned int Column) const
			noexcept
	{
	return Row < Height and Column < Width;
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Compute the index of a pixel in the bitmap, given the row and column
//!	indices.
//!
//! \param	Row		Row index.
//! \param	Column		Column index.
//!
//! \throw	None.
//!
//! \returns	IndexType	The index.
//!
//! \note
//!	If Row is not less that Height or Column is not less than Width, the
//!	returned value is not a valid index.
//!
////////////////////////////////////////////////////////////////////////////////

inline	Bitmap::IndexType	Bitmap::MakeIndex(unsigned int Row,
					unsigned int Column) const noexcept
	{
	assert(Row < Height);
	assert(Column < Width);
	return static_cast<IndexType>(Row) * Width + Column;
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Check whether two pixels are orthogonal at distance one to each other.
//!
//! \param	FirstIndex	Index of first pixel.
//! \param	SecondIndex	Index of second pixel.
//!
//! \throw	None.
//!
//! \returns	bool		true, if the two pixels are orthogonal at
//!				distance one to each other.
//!
//! \note
//!	If no image has been read, the returned value is undefined.
//!
////////////////////////////////////////////////////////////////////////////////

inline	bool	Bitmap::IsOrthogonal(IndexType FirstIndex,
			IndexType SecondIndex) const noexcept
	{
	if (FirstIndex >= SecondIndex)
		FirstIndex -= SecondIndex;
	else
		FirstIndex = SecondIndex - FirstIndex;
	return FirstIndex == 1 or FirstIndex == Width;
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Divide a uint_fast32_t by another, rounding the quotient.
//!
//! \param	Dividend	Dividend.
//! \param	Divisor		Divisor.
//!
//! \throw	None.
//!
//! \returns	uint_fast32_t	The rounded quotient.
//!
//! \note
//!	If Divisor is zero, the behavior is undefined.
//!
//! \note
//!	If Dividend + Divisor / 2 is greater than UINT_FAST32_MAX, the returned
//!	value is undefined.
//!
////////////////////////////////////////////////////////////////////////////////

inline	uint_fast32_t	Bitmap::RoundedDivide(uint_fast32_t Dividend,
				uint_fast32_t Divisor) noexcept
	{
	return (Dividend + Divisor / 2) / Divisor;
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Convert consecutive bytes in little endian order to an uint_fast32_t.
//!
//! \param	Buffer		Array of bytes to convert.
//! \param	Size		The number of bytes to convert.
//!
//! \throw	None.
//!
//! \returns	uint_fast32_t	The converted value.
//!
//! \note
//!	If Size is different from SHORT_DATA_SIZE or LONG_DATA_SIZE, the
//!	function returns zero.
//!
//! \note
//!	If Buffer is not a pointer to an area of Size bytes, the behavior is
//!	undefined.
//!
////////////////////////////////////////////////////////////////////////////////

inline	uint_fast32_t	Bitmap::LoadInteger(const uint8_t Buffer [],
				unsigned int Size) noexcept
	{
	switch (Size)
		{
		case SHORT_DATA_SIZE:
			return static_cast<uint_fast32_t>(Buffer [0]) bitor
				(static_cast<uint_fast32_t>(Buffer [1]) <<
				CHAR_BIT);
		case LONG_DATA_SIZE:
			return static_cast<uint_fast32_t>(Buffer [0]) bitor
				(static_cast<uint_fast32_t>(Buffer [1]) <<
				CHAR_BIT) bitor
				(static_cast<uint_fast32_t>(Buffer [2]) <<
				(CHAR_BIT * 2)) bitor
				(static_cast<uint_fast32_t>(Buffer [3]) <<
				(CHAR_BIT * 3));
		default:
			assert(false);
			return 0;
		}
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Convert an integer to consecutive bytes, in little endian order.
//!
//! \param	Buffer	Array where bytes will be stored.
//! \param	Value	Value to be stored.
//! \param	Size	The number of bytes to convert.
//!
//! \throw	None.
//!
//! \returns	None.
//!
//! \note
//!	If Size is different from SHORT_DATA_SIZE or LONG_DATA_SIZE, the
//!	behavior is undefined.
//!
//! \note
//!	If Buffer is not a pointer to an area of Size bytes, the behavior is
//!	undefined.
//!
////////////////////////////////////////////////////////////////////////////////

inline	void	Bitmap::StoreInteger(uint8_t Buffer [], uint_fast32_t Value,
			unsigned int Size) noexcept
	{
	assert(Size == SHORT_DATA_SIZE or Size == LONG_DATA_SIZE);
	Buffer [0] = static_cast<uint8_t>(Value);
	Buffer [1] = static_cast<uint8_t>(Value >> CHAR_BIT);
	if (Size == LONG_DATA_SIZE)
		{
		Buffer [2] = static_cast<uint8_t>(Value >> (CHAR_BIT * 2));
		Buffer [3] = static_cast<uint8_t>(Value >> (CHAR_BIT * 3));
		}
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Return the width of the image.
//!
//! \throw	None.
//!
//! \returns	uint_fast32_t	The height of the image.
//!
//! \note
//!	If no image has been read, the returned value is undefined.
//!
////////////////////////////////////////////////////////////////////////////////

inline	uint_fast32_t	Bitmap::GetWidth(void) const noexcept
	{
	return Width;
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Return the height of the image.
//!
//! \throw	None.
//!
//! \returns	uint_fast32_t	The height of the image.
//!
//! \note
//!	If no image has been read, the returned value is undefined.
//!
////////////////////////////////////////////////////////////////////////////////

inline	uint_fast32_t	Bitmap::GetHeight(void) const noexcept
	{
	return Height;
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Return the size of the image, in pixels,
//!
//! \throw	None.
//!
//! \returns	uint_fast32_t	The size of the image.
//!
//! \note
//!	If no image has been read, the returned value is undefined.
//!
////////////////////////////////////////////////////////////////////////////////

inline	uint_fast32_t	Bitmap::GetSize(void) const noexcept
	{
	return BitmapSize;
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Return the number of bits per pixel.
//!
//! \throw	None.
//!
//! \returns	unsigned int	The number of bits per pixel.
//!
//! \note
//!	If no image has been read, the returned value is undefined.
//!
////////////////////////////////////////////////////////////////////////////////

inline	unsigned int	Bitmap::GetBitsPerPixel(void) const noexcept
	{
	return BitsPerPixel;
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Return the horizontal resolution.
//!
//! \throw	None.
//!
//! \returns	uint_fast32_t	The horizontal resolution.
//!
//! \note
//!	If no image has been read, the returned value is undefined.
//!
////////////////////////////////////////////////////////////////////////////////

inline	uint_fast32_t	Bitmap::GetHorizontalResolution(void) const noexcept
	{
	return HorizontalResolution;
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Return the vertical resolution.
//!
//! \throw	None.
//!
//! \returns	uint_fast32_t	The vertical resolution.
//!
//! \note
//!	If no image has been read, the returned value is undefined.
//!
////////////////////////////////////////////////////////////////////////////////

inline	uint_fast32_t	Bitmap::GetVerticalResolution(void) const noexcept
	{
	return VerticalResolution;
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Compute the size of a line in bytes in a bitmap file in bytes.
//!
//! \throw	None.
//!
//! \returns	uint_fast32_t	The size of a line in a bitmap file.
//!
////////////////////////////////////////////////////////////////////////////////

inline	uint_fast32_t	Bitmap::ComputeLineSize(void) const noexcept
	{
	return (Width * BitsPerPixel + CHAR_BIT * sizeof(uint32_t) - 1) /
		(CHAR_BIT * sizeof(uint32_t)) * sizeof(uint32_t);
	}

////////////////////////////////////////////////////////////////////////////////
//! \par	Description:
//!	Initialize the random numbers generator with a set of values.
//!
//! \throw	None.
//!
//! \returns	None.
//!
//! \see	NUMBER_OF_GENERATORS
//! \see	INITIAL_GENERATION
//!
//! \note
//!	The function causes the generation of a small number of random numbers,
//!	to better initialize the generator.
//!
////////////////////////////////////////////////////////////////////////////////

inline	void	Bitmap::Randomize(uint_fast32_t Seed [RandomGenerator::
			NUMBER_OF_GENERATORS + 1]) noexcept
	{
	register	unsigned int	Index;

	Generator.SetSeed(Seed);
	for (Index = 0; Index < INITIAL_GENERATION; ++Index)
		IgnoreValue(Generator.Generate(2));
	}

}
